#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

"""This file was generated by xsdata, v23.8, on 2023-10-05 09:59:45

Generator: DataclassGenerator
See: https://xsdata.readthedocs.io/
"""
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union

from xsdata.models.datatype import XmlDate, XmlDateTime

from .external.ism.schema.cvegenerated.cvenum_ism25_x import CVEnumISM25X
from .external.ism.schema.cvegenerated.cvenum_ismclassification_all import CVEnumISMClassificationAll
from .external.ism.schema.cvegenerated.cvenum_ismcomplies_with import CVEnumISMCompliesWithValues
from .external.ism.schema.cvegenerated.cvenum_ismdissem import CVEnumISMDissemValuesvalue
from .external.ism.schema.cvegenerated.cvenum_ismfgiopen import CVEnumISMFGIOpenValues
from .external.ism.schema.cvegenerated.cvenum_ismfgiprotected import CVEnumISMFGIProtectedValues
from .external.ism.schema.cvegenerated.cvenum_ismnon_ic import CVEnumISMNonICValues
from .external.ism.schema.cvegenerated.cvenum_ismnon_uscontrols import CVEnumISMNonUSControlsValues
from .external.ism.schema.cvegenerated.cvenum_ismowner_producer import CVEnumISMOwnerProducerValues
from .external.ism.schema.cvegenerated.cvenum_ismrel_to import CVEnumISMRelToValues
from .external.ism.schema.cvegenerated.cvenum_ismscicontrols import CVEnumISMSCIControlsValuesvalue
from .external.ism.schema.cvegenerated.cvenum_ismsource_marked import CVEnumISMSourceMarked
from .sfa import (
    LinearRingType,
    LineType,
    MultiLineStringType,
    MultiPointType,
    MultiPolygonType,
    PointType,
    PolygonType,
    PolyhedralSurfaceType,
    ReferenceSystemType,
)
from .sicommon_types import (
    AngleMagnitudeType,
    ErrorStatisticsType,
    LatLonVertexType,
    ParameterType,
    PolarizationType,
    Poly2DType,
    RadarModeType,
    RadiometricType,
    RangeAzimuthType,
    ReferencePointType,
    RowColDoubleType,
    RowColIntType,
    XYZPolyType,
    XYZType,
)

__NAMESPACE__ = "urn:SIDD:1.0.0"


@dataclass
class AcheivedResolutionType:
    """
    Finest achievable resolution parameters.
    """


@dataclass
class ClassificationGuidanceType:
    """
    Classification guidance authority (only if file is classified).

    :ivar authority: Classifying authority.
    :ivar date: Date that the authority was provided. Specified in YYYY-MM-DD.
    """

    authority: Optional[str] = field(
        default=None,
        metadata={
            "name": "Authority",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    date: Optional[XmlDate] = field(
        default=None,
        metadata={
            "name": "Date",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class DRAHistogramOverridesType:
    """
    :ivar clip_min: Suggested override for the lower end-point of the display histogram in the ELT DRA
        application. Referred to as Pmin in SIPS documentation.
    :ivar clip_max: Suggested override for the upper end-point of the display histogram in the ELT DRA
        application. Referred to as Pmax in SIPS documentation.
    """

    clip_min: Optional[int] = field(
        default=None,
        metadata={
            "name": "ClipMin",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    clip_max: Optional[int] = field(
        default=None,
        metadata={
            "name": "ClipMax",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


class DecimationMethodType(Enum):
    """Default ELT decimation method for this data.

    Also used as default for reduced resolution dataset generation (if applicable).
    """

    NEAREST_NEIGHBOR = "NEAREST_NEIGHBOR"
    BILINEAR = "BILINEAR"
    BRIGHTEST_PIXEL = "BRIGHTEST_PIXEL"
    LAGRANGE = "LAGRANGE"


@dataclass
class Lookup3TableType:
    """
    :ivar value:
    :ivar size: Size of LUT
    """

    value: List[str] = field(
        default_factory=list,
        metadata={
            "pattern": r"([0-9]+),([0-9]+),([0-9]+)",
            "tokens": True,
        },
    )
    size: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )


@dataclass
class LookupTableType:
    """
    :ivar value:
    :ivar size: Size of LUT.
    """

    value: List[int] = field(
        default_factory=list,
        metadata={
            "tokens": True,
        },
    )
    size: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )


class MagnificationMethodType(Enum):
    """
    Default ELT magnification method for this data.
    """

    NEAREST_NEIGHBOR = "NEAREST_NEIGHBOR"
    BILINEAR = "BILINEAR"
    LAGRANGE = "LAGRANGE"


@dataclass
class MonitorCompensationAppliedType:
    """
    Describes monitor compensation that may have been applied to the product during
    processing.

    :ivar gamma: Gamma value for monitor compensation pre-applied to the image.
    :ivar xmin: Xmin value for monitor compensation pre-applied to the image.
    """

    gamma: Optional[float] = field(
        default=None,
        metadata={
            "name": "Gamma",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    xmin: Optional[float] = field(
        default=None,
        metadata={
            "name": "XMin",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


class PixelType(Enum):
    MONO8_I = "MONO8I"
    MONO8_LU = "MONO8LU"
    MONO16_I = "MONO16I"
    RGB8_LU = "RGB8LU"
    RGB24_I = "RGB24I"


@dataclass
class ProcessorInformationType:
    """
    :ivar application: Software application name and version number.
    :ivar processing_date_time: Date and time defined in Coordinated Universal Time (UTC). The seconds should be
        followed by a Z to indicate UTC.
    :ivar site: Creation location of product.
    :ivar profile: Product-specific profile applied during product processing.
    """

    application: Optional[str] = field(
        default=None,
        metadata={
            "name": "Application",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    processing_date_time: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "ProcessingDateTime",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    site: Optional[str] = field(
        default=None,
        metadata={
            "name": "Site",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    profile: Optional[str] = field(
        default=None,
        metadata={
            "name": "Profile",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class AnnotationObjectType:
    """
    Geometrical representation of the annotation.
    """

    point: Optional[PointType] = field(
        default=None,
        metadata={
            "name": "Point",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    line: Optional[LineType] = field(
        default=None,
        metadata={
            "name": "Line",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    linear_ring: Optional[LinearRingType] = field(
        default=None,
        metadata={
            "name": "LinearRing",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    polygon: Optional[PolygonType] = field(
        default=None,
        metadata={
            "name": "Polygon",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    polyhedral_surface: Optional[PolyhedralSurfaceType] = field(
        default=None,
        metadata={
            "name": "PolyhedralSurface",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    multi_polygon: Optional[MultiPolygonType] = field(
        default=None,
        metadata={
            "name": "MultiPolygon",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    multi_line_string: Optional[MultiLineStringType] = field(
        default=None,
        metadata={
            "name": "MultiLineString",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    multi_point: Optional[MultiPointType] = field(
        default=None,
        metadata={
            "name": "MultiPoint",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class BaseProjectionType:
    """
    :ivar reference_point: Reference point for the geometrical system.
    """

    reference_point: Optional[ReferencePointType] = field(
        default=None,
        metadata={
            "name": "ReferencePoint",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class ColorDisplayRemapType:
    """
    Object representing that the data requires color display.

    :ivar remap_lut: LUT-base remap indicating that the color display is done through index-based color.
    """

    remap_lut: Optional[Lookup3TableType] = field(
        default=None,
        metadata={
            "name": "RemapLUT",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class ExploitationFeaturesCollectionGeometryType:
    """
    Key geometry parameters independent of product processing.

    :ivar azimuth: Angle clockwise from north indicating the ETP line of sight vector.
    :ivar slope: Angle between the ETP at scene center and the range vector perpendicular to the direction of
        motion.
    :ivar squint: Angle from the ground track to platform velocity vector at nadir. Left-look is negative, right-
        look is positive.
    :ivar graze: Angle between the ETP and the line of sight vector.
    :ivar tilt: Angle between the ETP and the cross range vector. Also known as the twist angle.
    :ivar extension: Exploitation feature extension related to geometry for a single input image
    """

    azimuth: Optional[float] = field(
        default=None,
        metadata={
            "name": "Azimuth",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    slope: Optional[float] = field(
        default=None,
        metadata={
            "name": "Slope",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": 0.0,
            "max_inclusive": 90.0,
        },
    )
    squint: Optional[float] = field(
        default=None,
        metadata={
            "name": "Squint",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": -180.0,
            "max_inclusive": 180.0,
        },
    )
    graze: Optional[float] = field(
        default=None,
        metadata={
            "name": "Graze",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": 0.0,
            "max_inclusive": 90.0,
        },
    )
    tilt: Optional[float] = field(
        default=None,
        metadata={
            "name": "Tilt",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": -180.0,
            "max_inclusive": 180.0,
        },
    )
    extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "Extension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ExploitationFeaturesCollectionPhenomenologyType:
    """
    Phenomenology related to both the geometry and the final product processing.

    :ivar shadow: The phenomon where vertical objects occlude radar energy.
    :ivar layover: The phenomenon where vertical objects appear as ground objects with the same range/range rate.
    :ivar multi_path: This is a range dependent phenomenon which describes the energy from a single scatter
        returned to the radar via more than one path and results in a nominally constant direction in the ETP.
    :ivar ground_track: Counter-clockwise angle from increasing row direction to ground track at the center of
        the image.
    :ivar extension: Exploitation feature extension related to phenomenology for a single input image
    """

    shadow: Optional[AngleMagnitudeType] = field(
        default=None,
        metadata={
            "name": "Shadow",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    layover: Optional[AngleMagnitudeType] = field(
        default=None,
        metadata={
            "name": "Layover",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    multi_path: Optional[float] = field(
        default=None,
        metadata={
            "name": "MultiPath",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": -180.0,
            "max_inclusive": 180.0,
        },
    )
    ground_track: Optional[float] = field(
        default=None,
        metadata={
            "name": "GroundTrack",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": -180.0,
            "max_inclusive": 180.0,
        },
    )
    extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "Extension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ExploitationFeaturesProductType:
    """
    Metadata regarding the product.

    :ivar resolution: Uniformly-weighted resolution projected into the Earth Tangent Plane (ETP).
    :ivar north: Counter-clockwise angle from increasing row direction to north at the center of the image.
    :ivar extension: Exploitation feature extension for the end product
    """

    resolution: Optional[RowColDoubleType] = field(
        default=None,
        metadata={
            "name": "Resolution",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    north: Optional[float] = field(
        default=None,
        metadata={
            "name": "North",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": -180.0,
            "max_inclusive": 180.0,
        },
    )
    extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "Extension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class FootprintType:
    vertex: List[LatLonVertexType] = field(
        default_factory=list,
        metadata={
            "name": "Vertex",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_occurs": 4,
            "max_occurs": 4,
        },
    )
    size: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )


@dataclass
class GeographicInformationType:
    """
    :ivar country_code: Country identifier for this geographic region.
    :ivar security_info: Specifies classification level or special handling designators for this geographic
        region
    :ivar geographic_info_extension: Implementation specific geographic information.
    """

    country_code: List[str] = field(
        default_factory=list,
        metadata={
            "name": "CountryCode",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    security_info: Optional[str] = field(
        default=None,
        metadata={
            "name": "SecurityInfo",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    geographic_info_extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "GeographicInfoExtension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class GeometricChipType:
    """
    :ivar chip_size: Size of the chipped product in pixels.
    :ivar original_upper_left_coordinate: Upper-left corner with respect to the original product.
    :ivar original_upper_right_coordinate: Upper-right corner with respect to the original product.
    :ivar original_lower_left_coordinate: Lower-left corner with respect to the original product.
    :ivar original_lower_right_coordinate: Lower-right corner with respect to the original product.
    """

    chip_size: Optional[RowColIntType] = field(
        default=None,
        metadata={
            "name": "ChipSize",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    original_upper_left_coordinate: Optional[RowColDoubleType] = field(
        default=None,
        metadata={
            "name": "OriginalUpperLeftCoordinate",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    original_upper_right_coordinate: Optional[RowColDoubleType] = field(
        default=None,
        metadata={
            "name": "OriginalUpperRightCoordinate",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    original_lower_left_coordinate: Optional[RowColDoubleType] = field(
        default=None,
        metadata={
            "name": "OriginalLowerLeftCoordinate",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    original_lower_right_coordinate: Optional[RowColDoubleType] = field(
        default=None,
        metadata={
            "name": "OriginalLowerRightCoordinate",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class InputROIType:
    """
    ROI representing portion of input data used to make this product.

    :ivar size: Number of rows and columns extracted from the input.
    :ivar upper_left: The upper-left pixel extracted from the input.
    """

    size: Optional[RowColIntType] = field(
        default=None,
        metadata={
            "name": "Size",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    upper_left: Optional[RowColIntType] = field(
        default=None,
        metadata={
            "name": "UpperLeft",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class MonochromeDisplayRemapType:
    """This remap works by taking the input space and using the LUT to map it to a
    log space (for 8-bit only).

    From the log space the C0 and Ch fields are applied to get to display-ready density space.
    The density should then be rendered by the TTC and monitor comp.
    This means that the default DRA should not apply anything besides the clip points.
    If a different contrast/brightness is applied it should be done through modification of the clip points via DRA.
    Examples:
    Remap                LUT                    Clips
    =============================
    PEDF            PEDF-&gt;D                0,255
    LLG            LLG-&gt;A-&gt;LogA        C0,Ch
    Log                N/A                         C0,Ch
    NRL                N/A                        0,255 (Supposed to be display ready)

    :ivar remap_type: Name of remap applied (for informational purposes only).
    :ivar remap_lut: Lookup table for remap to log amplitude for display.  Used during the "Product Generation
        Option" portion of the SIPS display chain. Required for 8-bit data. Not to be used for 16-bit data.
    :ivar remap_parameter: Textual remap parameter. Filled based upon remap type (for informational purposes
        only).  For example, if the data is linlog encoded a RemapParameter could be used to describe any
        amplitude scaling that was performed prior to linlog encoding the data.
    """

    remap_type: Optional[str] = field(
        default=None,
        metadata={
            "name": "RemapType",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    remap_lut: Optional[LookupTableType] = field(
        default=None,
        metadata={
            "name": "RemapLUT",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    remap_parameter: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "RemapParameter",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ProcessingEventType:
    """
    :ivar application_name: Application which applied a modification.
    :ivar applied_date_time: Date and time defined in Coordinated Universal Time (UTC). The seconds should be
        followed by a Z to indicate UTC.
    :ivar interpolation_method: Type of interpolation applied to the data.
    :ivar descriptor: Descriptor for the processing event.
    """

    application_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "ApplicationName",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    applied_date_time: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "AppliedDateTime",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    interpolation_method: Optional[str] = field(
        default=None,
        metadata={
            "name": "InterpolationMethod",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    descriptor: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "Descriptor",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ProcessingModuleType:
    """
    :ivar module_name: The name of the algorithm used in processing the product.
    :ivar module_parameter: Parameters associated with the algorithm used in processing the product.
    :ivar processing_module: ProcessingModule is a repeatable structure within itself to create an algorithm as a
        subset of another algorithm.
    """

    module_name: Optional[ParameterType] = field(
        default=None,
        metadata={
            "name": "ModuleName",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    module_parameter: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "ModuleParameter",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    processing_module: List["ProcessingModuleType"] = field(
        default_factory=list,
        metadata={
            "name": "ProcessingModule",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ProductClassificationType:
    """
    The overall classification of the product.

    :ivar security_extension: Extensible parameters used to support profile-specific needs related to product
        security.
    :ivar desversion: The version number of the DES. Should there be multiple specified in an instance document
        the one at the root node is the one that will apply to the entire document.
    :ivar resource_element:
    :ivar create_date:
    :ivar complies_with:
    :ivar classification:
    :ivar owner_producer:
    :ivar scicontrols:
    :ivar saridentifier:
    :ivar dissemination_controls:
    :ivar fgisource_open:
    :ivar fgisource_protected:
    :ivar releasable_to:
    :ivar non_icmarkings:
    :ivar classified_by:
    :ivar compilation_reason:
    :ivar derivatively_classified_by:
    :ivar classification_reason:
    :ivar non_uscontrols:
    :ivar derived_from:
    :ivar declass_date:
    :ivar declass_event:
    :ivar declass_exception:
    :ivar type_of_exempted_source:
    :ivar date_of_exempted_source:
    """

    security_extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "SecurityExtension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    desversion: int = field(
        init=False,
        default=4,
        metadata={
            "name": "DESVersion",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "required": True,
        },
    )
    resource_element: bool = field(
        init=False,
        default=True,
        metadata={
            "name": "resourceElement",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "required": True,
        },
    )
    create_date: Optional[XmlDate] = field(
        default=None,
        metadata={
            "name": "createDate",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "required": True,
        },
    )
    complies_with: List[CVEnumISMCompliesWithValues] = field(
        default_factory=list,
        metadata={
            "name": "compliesWith",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "tokens": True,
        },
    )
    classification: Optional[CVEnumISMClassificationAll] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "required": True,
        },
    )
    owner_producer: List[CVEnumISMOwnerProducerValues] = field(
        default_factory=list,
        metadata={
            "name": "ownerProducer",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "required": True,
            "tokens": True,
        },
    )
    scicontrols: List[Union[str, CVEnumISMSCIControlsValuesvalue]] = field(
        default_factory=list,
        metadata={
            "name": "SCIcontrols",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "pattern": r"SI-G-[A-Z][A-Z][A-Z][A-Z]|SI-ECI-[A-Z][A-Z][A-Z]",
            "tokens": True,
        },
    )
    saridentifier: List[str] = field(
        default_factory=list,
        metadata={
            "name": "SARIdentifier",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "pattern": r"SAR-[A-Z][A-Z][A-Z]?",
            "tokens": True,
        },
    )
    dissemination_controls: List[Union[str, CVEnumISMDissemValuesvalue]] = field(
        default_factory=list,
        metadata={
            "name": "disseminationControls",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "pattern": r"RD-SG-[1-9][0-9]?|FRD-SG-[1-9][0-9]?",
            "tokens": True,
        },
    )
    fgisource_open: List[CVEnumISMFGIOpenValues] = field(
        default_factory=list,
        metadata={
            "name": "FGIsourceOpen",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "tokens": True,
        },
    )
    fgisource_protected: List[CVEnumISMFGIProtectedValues] = field(
        default_factory=list,
        metadata={
            "name": "FGIsourceProtected",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "tokens": True,
        },
    )
    releasable_to: List[CVEnumISMRelToValues] = field(
        default_factory=list,
        metadata={
            "name": "releasableTo",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "tokens": True,
        },
    )
    non_icmarkings: List[CVEnumISMNonICValues] = field(
        default_factory=list,
        metadata={
            "name": "nonICmarkings",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "tokens": True,
        },
    )
    classified_by: Optional[str] = field(
        default=None,
        metadata={
            "name": "classifiedBy",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "max_length": 1024,
        },
    )
    compilation_reason: Optional[str] = field(
        default=None,
        metadata={
            "name": "compilationReason",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "max_length": 1024,
        },
    )
    derivatively_classified_by: Optional[str] = field(
        default=None,
        metadata={
            "name": "derivativelyClassifiedBy",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "max_length": 1024,
        },
    )
    classification_reason: Optional[str] = field(
        default=None,
        metadata={
            "name": "classificationReason",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "max_length": 4096,
        },
    )
    non_uscontrols: List[CVEnumISMNonUSControlsValues] = field(
        default_factory=list,
        metadata={
            "name": "nonUSControls",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "tokens": True,
        },
    )
    derived_from: Optional[str] = field(
        default=None,
        metadata={
            "name": "derivedFrom",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "max_length": 1024,
        },
    )
    declass_date: Optional[XmlDate] = field(
        default=None,
        metadata={
            "name": "declassDate",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
        },
    )
    declass_event: Optional[str] = field(
        default=None,
        metadata={
            "name": "declassEvent",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
            "max_length": 1024,
        },
    )
    declass_exception: Optional[CVEnumISM25X] = field(
        default=None,
        metadata={
            "name": "declassException",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
        },
    )
    type_of_exempted_source: Optional[CVEnumISMSourceMarked] = field(
        default=None,
        metadata={
            "name": "typeOfExemptedSource",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
        },
    )
    date_of_exempted_source: Optional[XmlDate] = field(
        default=None,
        metadata={
            "name": "dateOfExemptedSource",
            "type": "Attribute",
            "namespace": "urn:us:gov:ic:ism",
        },
    )


@dataclass
class ProductPlaneType:
    """
    Plane definition for the product.

    :ivar row_unit_vector: Unit vector of the plane defined to be aligned in the increasing row direction of the
        product. (Defined as Rpgd in Design and Exploitation document)
    :ivar col_unit_vector: Unit vector of the plane defined to be aligned in the increasing column direction of
        the product. (Defined as Cpgd in Design and Exploitation document)
    """

    row_unit_vector: Optional[XYZType] = field(
        default=None,
        metadata={
            "name": "RowUnitVector",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    col_unit_vector: Optional[XYZType] = field(
        default=None,
        metadata={
            "name": "ColUnitVector",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class TxRcvPolarizationType:
    """
    :ivar tx_polarization: Polarization transmit type
    :ivar rcv_polarization: Receive polarization type
    :ivar rcv_polarization_offset: Optional angle offset for the receive polarization defined at aperture center.
    :ivar processed: Optional flag to describe whether this input polarization was used in processing the
        product.
    """

    tx_polarization: Optional[PolarizationType] = field(
        default=None,
        metadata={
            "name": "TxPolarization",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    rcv_polarization: Optional[PolarizationType] = field(
        default=None,
        metadata={
            "name": "RcvPolarization",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    rcv_polarization_offset: Optional[float] = field(
        default=None,
        metadata={
            "name": "RcvPolarizationOffset",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_inclusive": -180.0,
            "max_inclusive": 180.0,
        },
    )
    processed: Optional[bool] = field(
        default=None,
        metadata={
            "name": "Processed",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class AnnotationType:
    """
    Single annotation.

    :ivar identifier: Identifier for the annotation which idicates the type of object represented by this
        annotation.
    :ivar spatial_reference_system: Spatial reference system of the annotation.  Assumed to be WGS-84 geographic
        coordinate system if not specified with (lat, lon, h) units in (arc-sec, arc-sec, meters above
        ellipsoid).
    :ivar object_value: The geometrical representation of the annotation.
    """

    identifier: Optional[str] = field(
        default=None,
        metadata={
            "name": "Identifier",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    spatial_reference_system: Optional[ReferenceSystemType] = field(
        default=None,
        metadata={
            "name": "SpatialReferenceSystem",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    object_value: List[AnnotationObjectType] = field(
        default_factory=list,
        metadata={
            "name": "Object",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_occurs": 1,
        },
    )


@dataclass
class DownstreamReprocessingType:
    """
    :ivar geometric_chip: Contains information related to downstream chipping of the product.
    :ivar processing_event: Contains information related to downstream processing of the product.
    """

    geometric_chip: Optional[GeometricChipType] = field(
        default=None,
        metadata={
            "name": "GeometricChip",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    processing_event: List[ProcessingEventType] = field(
        default_factory=list,
        metadata={
            "name": "ProcessingEvent",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ExploitationFeaturesCollectionInformationType:
    """
    General collection information.

    :ivar sensor_name: The name of the sensor.
    :ivar radar_mode: Radar collection mode.  The ModeType refers to the collection type [SPOTLIGHT, STRIPMAP,
        DYNAMIC STRIPMAP].  The optional ModeID is used to represent system-specific mode identifiers.
    :ivar collection_date_time: Collection date and time defined in Coordinated Universal Time (UTC). The seconds
        should be followed by a Z to indicate UTC.
    :ivar local_date_time: Date and time defined in local time.
    :ivar collection_duration: The duration of the collection (units = seconds).
    :ivar resolution: Uniformly-weighted resolution (range and azimuth) processed in the slant plane.
    :ivar input_roi: ROI representing portion of input data used to make this product.
    :ivar polarization: Transmit and receive polarization.
    """

    sensor_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "SensorName",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    radar_mode: Optional[RadarModeType] = field(
        default=None,
        metadata={
            "name": "RadarMode",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    collection_date_time: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "CollectionDateTime",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    local_date_time: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "name": "LocalDateTime",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    collection_duration: Optional[float] = field(
        default=None,
        metadata={
            "name": "CollectionDuration",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    resolution: Optional[RangeAzimuthType] = field(
        default=None,
        metadata={
            "name": "Resolution",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    input_roi: Optional[InputROIType] = field(
        default=None,
        metadata={
            "name": "InputROI",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    polarization: List[TxRcvPolarizationType] = field(
        default_factory=list,
        metadata={
            "name": "Polarization",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class GeographicCoverageType:
    """
    :ivar georegion_identifier: Identifier for the georegion.
    :ivar footprint: Estimated ground footprint of the product.
    :ivar sub_region: Used to represent hierarchical decomposition into sub-regions.
    :ivar geographic_info: Specifics about the georegion.
    """

    georegion_identifier: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "GeoregionIdentifier",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    footprint: Optional[FootprintType] = field(
        default=None,
        metadata={
            "name": "Footprint",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    sub_region: List["GeographicCoverageType"] = field(
        default_factory=list,
        metadata={
            "name": "SubRegion",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    geographic_info: Optional[GeographicInformationType] = field(
        default=None,
        metadata={
            "name": "GeographicInfo",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class MeasurableProjectionType(BaseProjectionType):
    """
    :ivar sample_spacing: Sample spacing in row and column.
    :ivar time_coapoly: Time (units = seconds) at which center of aperture for a given pixel coordinate in the
        product occurs.
    """

    sample_spacing: Optional[RowColDoubleType] = field(
        default=None,
        metadata={
            "name": "SampleSpacing",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    time_coapoly: Optional[Poly2DType] = field(
        default=None,
        metadata={
            "name": "TimeCOAPoly",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class PolynomialProjectionType(BaseProjectionType):
    """Polynomial pixel to ground.

    Only used for sensor systems where the radar geometry parameters are not recorded.

    :ivar row_col_to_lat: Polynomial that converts Row/Col to Latitude (degrees).
    :ivar row_col_to_lon: Polynomial that converts Row/Col to Longitude (degrees).
    :ivar row_col_to_alt: Polynomial that converts Row/Col to Altitude (meters above WGS-84 ellipsoid).
    :ivar lat_lon_to_row: Polynomial that converts Latitude (degrees) and Longitude (degrees) to pixel row
        location.
    :ivar lat_lon_to_col: Polynomial that converts Latitude (degrees) and Longitude (degrees) to pixel column
        location
    """

    row_col_to_lat: Optional[Poly2DType] = field(
        default=None,
        metadata={
            "name": "RowColToLat",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    row_col_to_lon: Optional[Poly2DType] = field(
        default=None,
        metadata={
            "name": "RowColToLon",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    row_col_to_alt: Optional[Poly2DType] = field(
        default=None,
        metadata={
            "name": "RowColToAlt",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    lat_lon_to_row: Optional[Poly2DType] = field(
        default=None,
        metadata={
            "name": "LatLonToRow",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    lat_lon_to_col: Optional[Poly2DType] = field(
        default=None,
        metadata={
            "name": "LatLonToCol",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class ProductCreationType:
    """
    Contains general information about product creation.

    :ivar processor_information: Details regarding processor.
    :ivar classification: The overall classification of the product.
    :ivar product_name: The output product name defined by the processor.
    :ivar product_class: Class of product. (examples: Dynamic Image, Amplitude Change Detection, Coherent Change
        Detection, etc.).
    :ivar product_type: Type of sub-product. (examples: Frame #, Reference, Match, etc.). This field is only
        needed if there is a suite of associated products.
    :ivar product_creation_extension: Extensible parameters used to support profile-specific needs related to
        product creation.
    """

    processor_information: Optional[ProcessorInformationType] = field(
        default=None,
        metadata={
            "name": "ProcessorInformation",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    classification: Optional[ProductClassificationType] = field(
        default=None,
        metadata={
            "name": "Classification",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    product_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "ProductName",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    product_class: Optional[str] = field(
        default=None,
        metadata={
            "name": "ProductClass",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    product_type: Optional[str] = field(
        default=None,
        metadata={
            "name": "ProductType",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    product_creation_extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "ProductCreationExtension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ProductProcessingType:
    """
    :ivar processing_module: Processing module to keep track of the name and any parameters associated with the
        algorithms used to produce the SIDD.
    """

    processing_module: List[ProcessingModuleType] = field(
        default_factory=list,
        metadata={
            "name": "ProcessingModule",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_occurs": 1,
        },
    )


@dataclass
class RemapChoiceType:
    """
    :ivar color_display_remap: Information for proper color display of the data.
    :ivar monochrome_display_remap: Information for proper monochrome display of the data.
    """

    color_display_remap: Optional[ColorDisplayRemapType] = field(
        default=None,
        metadata={
            "name": "ColorDisplayRemap",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    monochrome_display_remap: Optional[MonochromeDisplayRemapType] = field(
        default=None,
        metadata={
            "name": "MonochromeDisplayRemap",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class TargetInformationType:
    """
    :ivar identifier: Target may have one or more identifiers.  Examples: names, BE numbers, etc. Use the "name"
        attribute to describe what this is.
    :ivar footprint: Target footprint as defined by polygonal shape.
    :ivar target_information_extension: Generic extension.  Could be used to indicate type of target, terrain,
        etc.
    """

    identifier: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "Identifier",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_occurs": 1,
        },
    )
    footprint: Optional[FootprintType] = field(
        default=None,
        metadata={
            "name": "Footprint",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    target_information_extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "TargetInformationExtension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class AnnotationsType:
    """
    :ivar annotation: Annotation Object.
    """

    annotation: List[AnnotationType] = field(
        default_factory=list,
        metadata={
            "name": "Annotation",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_occurs": 1,
        },
    )


@dataclass
class CylindricalProjectionType(MeasurableProjectionType):
    """
    Cylindrical mapping of the pixel grid.

    :ivar stripmap_direction: Along stripmap direction
    :ivar curvature_radius: Radius of Curvature defined at scene center.  If not present, the radius of curvature
        will be derived based upon the equations provided in the Design and Exploitation Document
    """

    stripmap_direction: Optional[XYZType] = field(
        default=None,
        metadata={
            "name": "StripmapDirection",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    curvature_radius: Optional[float] = field(
        default=None,
        metadata={
            "name": "CurvatureRadius",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ExploitationFeaturesCollectionType:
    """
    :ivar information: General collection information.
    :ivar geometry: Key geometry parameters independent of product processing.
    :ivar phenomenology: Phenomenology related to both the geometry and the final product processing.
    """

    information: Optional[ExploitationFeaturesCollectionInformationType] = field(
        default=None,
        metadata={
            "name": "Information",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    geometry: Optional[ExploitationFeaturesCollectionGeometryType] = field(
        default=None,
        metadata={
            "name": "Geometry",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    phenomenology: Optional[ExploitationFeaturesCollectionPhenomenologyType] = field(
        default=None,
        metadata={
            "name": "Phenomenology",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class GeographicAndTargetType:
    """
    :ivar geographic_coverage: Provides geographic coverage information.
    :ivar target_information: Provides target specific geographic information.
    """

    geographic_coverage: Optional[GeographicCoverageType] = field(
        default=None,
        metadata={
            "name": "GeographicCoverage",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    target_information: List[TargetInformationType] = field(
        default_factory=list,
        metadata={
            "name": "TargetInformation",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class GeographicProjectionType(MeasurableProjectionType):
    """
    Geographic mapping of the pixel grid.
    """


@dataclass
class PlaneProjectionType(MeasurableProjectionType):
    """
    Planar representation of the pixel grid.

    :ivar product_plane: Plane definition for the product.
    """

    product_plane: Optional[ProductPlaneType] = field(
        default=None,
        metadata={
            "name": "ProductPlane",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class ProductDisplayType:
    """
    Type for describing proper display of the derived product.

    :ivar pixel_type: Defines the pixel type, based on enumeration and definition in Design and Exploitation
        document.
    :ivar remap_information: Information regarding the encoding of the pixel data. Used for 8-bit pixel types.
    :ivar magnification_method: Recommended ELT magnification method for this data.
    :ivar decimation_method: Recommended ELT decimation method for this data. Also used as default for reduced
        resolution dataset generation (if applicable).
    :ivar drahistogram_overrides: Recommended ELT DRA overrides.
    :ivar monitor_compensation_applied: Describes monitor compensation that may have been applied to the product
        during processing.
    :ivar display_extension: Extensible parameters used to support profile-specific needs related to product
        display.
    """

    pixel_type: Optional[PixelType] = field(
        default=None,
        metadata={
            "name": "PixelType",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    remap_information: Optional[RemapChoiceType] = field(
        default=None,
        metadata={
            "name": "RemapInformation",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    magnification_method: Optional[MagnificationMethodType] = field(
        default=None,
        metadata={
            "name": "MagnificationMethod",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    decimation_method: Optional[DecimationMethodType] = field(
        default=None,
        metadata={
            "name": "DecimationMethod",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    drahistogram_overrides: Optional[DRAHistogramOverridesType] = field(
        default=None,
        metadata={
            "name": "DRAHistogramOverrides",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    monitor_compensation_applied: Optional[MonitorCompensationAppliedType] = field(
        default=None,
        metadata={
            "name": "MonitorCompensationApplied",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    display_extension: List[ParameterType] = field(
        default_factory=list,
        metadata={
            "name": "DisplayExtension",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )


@dataclass
class ExploitationFeaturesType:
    """
    Computed metadata regarding the collect.

    :ivar collection: Metadata regarding one of the input collections.
    :ivar product: Metadata regarding the product.
    """

    collection: List["ExploitationFeaturesType.Collection"] = field(
        default_factory=list,
        metadata={
            "name": "Collection",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "min_occurs": 1,
        },
    )
    product: Optional[ExploitationFeaturesProductType] = field(
        default=None,
        metadata={
            "name": "Product",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )

    @dataclass
    class Collection(ExploitationFeaturesCollectionType):
        identifier: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )


@dataclass
class MeasurementType:
    """
    Geometric SAR information required for measurement/geolocation.

    :ivar polynomial_projection: Polynomial pixel to ground.  Only used for sensor systems where the radar
        geometry parameters are not recorded.
    :ivar geographic_projection: Geographic mapping of the pixel grid referred to as GGD in the Design and
        Exploitation document.
    :ivar plane_projection: Planar representation of the pixel grid referred to as PGD in the Design and
        Exploitation document.
    :ivar cylindrical_projection: Cylindrical mapping of the pixel grid referred to as CGD in the Design and
        Exploitation document.
    :ivar pixel_footprint: Size of the image.
    :ivar arppoly: Center of aperture polynomial (units = m) based upon time into the collect.
    """

    polynomial_projection: Optional[PolynomialProjectionType] = field(
        default=None,
        metadata={
            "name": "PolynomialProjection",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    geographic_projection: Optional[GeographicProjectionType] = field(
        default=None,
        metadata={
            "name": "GeographicProjection",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    plane_projection: Optional[PlaneProjectionType] = field(
        default=None,
        metadata={
            "name": "PlaneProjection",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    cylindrical_projection: Optional[CylindricalProjectionType] = field(
        default=None,
        metadata={
            "name": "CylindricalProjection",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
        },
    )
    pixel_footprint: Optional[RowColIntType] = field(
        default=None,
        metadata={
            "name": "PixelFootprint",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )
    arppoly: Optional[XYZPolyType] = field(
        default=None,
        metadata={
            "name": "ARPPoly",
            "type": "Element",
            "namespace": "urn:SIDD:1.0.0",
            "required": True,
        },
    )


@dataclass
class SIDD:
    """
    Root element of the SIDD document.

    :ivar product_creation: Information related to processor, classification, and product type.
    :ivar display: Contains information on the parameters needed to display the product in an exploitation tool.
    :ivar geographic_and_target: Contains generic and extensible targeting and geographic region information.
    :ivar measurement: Contains the metadata necessary for performing measurements.
    :ivar exploitation_features: Computed metadata regarding the input collections and final product.
    :ivar product_processing: Contains metadata related to algorithms used during product generation.
    :ivar downstream_reprocessing: Contains metadata related to downstream processing of the product.
    :ivar error_statistics: See SICD documentation for metadata definitions.
    :ivar radiometric: Radiometric information about the product.
    :ivar annotations: List of annotations for the imagery.
    """

    class Meta:
        namespace = "urn:SIDD:1.0.0"

    product_creation: Optional[ProductCreationType] = field(
        default=None,
        metadata={
            "name": "ProductCreation",
            "type": "Element",
            "required": True,
        },
    )
    display: Optional[ProductDisplayType] = field(
        default=None,
        metadata={
            "name": "Display",
            "type": "Element",
            "required": True,
        },
    )
    geographic_and_target: Optional[GeographicAndTargetType] = field(
        default=None,
        metadata={
            "name": "GeographicAndTarget",
            "type": "Element",
            "required": True,
        },
    )
    measurement: Optional[MeasurementType] = field(
        default=None,
        metadata={
            "name": "Measurement",
            "type": "Element",
            "required": True,
        },
    )
    exploitation_features: Optional[ExploitationFeaturesType] = field(
        default=None,
        metadata={
            "name": "ExploitationFeatures",
            "type": "Element",
            "required": True,
        },
    )
    product_processing: Optional[ProductProcessingType] = field(
        default=None,
        metadata={
            "name": "ProductProcessing",
            "type": "Element",
        },
    )
    downstream_reprocessing: Optional[DownstreamReprocessingType] = field(
        default=None,
        metadata={
            "name": "DownstreamReprocessing",
            "type": "Element",
        },
    )
    error_statistics: Optional[ErrorStatisticsType] = field(
        default=None,
        metadata={
            "name": "ErrorStatistics",
            "type": "Element",
        },
    )
    radiometric: Optional[RadiometricType] = field(
        default=None,
        metadata={
            "name": "Radiometric",
            "type": "Element",
        },
    )
    annotations: Optional[AnnotationsType] = field(
        default=None,
        metadata={
            "name": "Annotations",
            "type": "Element",
        },
    )



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

"""This file was generated by xsdata, v23.8, on 2023-10-05 09:59:45

Generator: DataclassGenerator
See: https://xsdata.readthedocs.io/
"""
from enum import Enum

__NAMESPACE__ = "urn:us:gov:ic:ism-cvenum"


class CVEnumISM25X(Enum):
    """(U) All currently authorized 25X values.

    PERMISSIBLE VALUES
    The permissible values for this simple type are defined in the Controlled Value Enumeration:
    CVEnumISM25X.xml

    :cvar VALUE_25_X1: Reveal information about the application of an intelligence source or method.
    :cvar VALUE_25_X1_HUMAN: Reveal the identity of a confidential human source or human intelligence source.
    :cvar VALUE_25_X2: Reveal information that would assist in the development or use of weapons of mass
        destruction.
    :cvar VALUE_25_X3: Reveal information that would impair U.S. cryptologic systems or activities.
    :cvar VALUE_25_X4: Reveal information that would impair the application of state-of-the-art technology within
        a U.S. weapon system.
    :cvar VALUE_25_X5: Reveal actual U.S. military war plans that remain in effect.
    :cvar VALUE_25_X6: Reveal information, including foreign government information, that would seriously and
        demonstrably impair relations between the United States and a foreign government or seriously and
        demonstrably undermine ongoing diplomatic activities of the United States.
    :cvar VALUE_25_X7: Reveal information that would clearly and demonstrably impair the current ability of
        United States Government officials to protect the President, Vice President, or other protectees for whom
        protection services, in the interest of national security, are authorized.
    :cvar VALUE_25_X8: Reveal information that would seriously and demonstrably impair current national security
        emergency preparedness plans or reveal current vulnerabilities of systems, installations,
        infrastructures, or projects relating to the national security.
    :cvar VALUE_25_X9: Violate a statue, treaty, or international agreement.
    """

    VALUE_25_X1 = "25X1"
    VALUE_25_X1_HUMAN = "25X1-human"
    VALUE_25_X2 = "25X2"
    VALUE_25_X3 = "25X3"
    VALUE_25_X4 = "25X4"
    VALUE_25_X5 = "25X5"
    VALUE_25_X6 = "25X6"
    VALUE_25_X7 = "25X7"
    VALUE_25_X8 = "25X8"
    VALUE_25_X9 = "25X9"



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import base64
import re
from io import StringIO
from typing import Callable, List, Optional, TypeVar
from xml.etree import ElementTree as ET

from defusedxml import ElementTree

# This is a type placeholder needed by the _get_tre_field_value() type hints
T = TypeVar("T")


class NITFDESAccessor:
    """
    This class is a facade that makes it easier to work with the XML formatted Data Extension Segments parsed by
    GDAL.
    """

    def __init__(self, gdal_xml_des_metadata: List[str]):
        """
        Construct the DES accessor given the contents of the xml:DES metadata domain from GDAL.
        ``
            raster_dataset = gdal.Open(path_to_nitf_image)
            xml_des = raster_dataset.GetMetadata("xml:DES")
            des_accessor = NITFDESAccessor(xml_des)
        ``

        :param gdal_xml_des_metadata: the GDAL parsed DES metadata
        """
        self.parsed_des_lists = []
        if gdal_xml_des_metadata is not None and len(gdal_xml_des_metadata) > 0:
            for xml_des_list in gdal_xml_des_metadata:
                # The new handling GDAL has for XML data content causes an XML document to be expanded in the middle
                # of the xml:DES data structure. An embedded xml prolog (e.g. <?xml version= ... ?>) is invalid syntax
                # that will throw off some XML parsers. The xml prolog is optional, so we can strip all of them from
                # the XML as a workaround while we look for a better way to address this recent GDAL change.
                clean_xml_string = re.sub(r"<\?xml.*?\?>", "", xml_des_list).strip()
                des_list = ElementTree.fromstring(clean_xml_string)
                self.parsed_des_lists.append(des_list)

    def get_segments_by_name(self, des_name: str) -> List[ET.Element]:
        """
        This method searches through the GDAL xml:DES metadata and returns the XML structure for any segments
        matching the provided name. This is equivalent to retrieving all segments that have a matching DESID.

        :param des_name: the DESID (e.g. XML_DATA_CONTENT)
        :return: the list of segments, multiple items in the list will occur if the NITF has multiple matching segments
        """
        result = []
        for des_list in self.parsed_des_lists:
            result.extend(des_list.findall(f"./des[@name='{des_name}']"))
        return result

    @staticmethod
    def extract_des_header(des_element: ET.Element) -> str:
        """
        This function encodes the existing values from the Data Extension Segment header into a fields appropriately
        sized for including in a NITF image. The DESDATA field is not copied because the assumption is that the data
        itself will be updated. The DE and DESID fields are not included either because GDAL adds them itself
        when writing the segment.

        :param des_element: the Data Extension Segment containing current segment
        :return: the encoded DESVER through DESSHF fields.
        """
        result_builder = StringIO()
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESVER", int), "02d"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DECLAS", str), ">1"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCLSY", str), ">2"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCODE", str), ">11"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCTLH", str), ">2"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESREL", str), ">20"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESDCTP", str), ">2"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESDCDT", str), ">8"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESDCXM", str), ">4"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESDG", str), ">1"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESDGDT", str), ">8"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCLTX", str), ">43"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCATP", str), ">1"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCAUT", str), ">40"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCRSN", str), ">1"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESSRDT", str), ">8"))
        result_builder.write(format(NITFDESAccessor.parse_field_value(des_element, "DESCTLN", str), ">15"))

        # TODO: If DESID = TRE_OVERFLOW DESOFLW, DESITEM

        subheader_length = NITFDESAccessor.parse_field_value(des_element, "DESSHL", int)
        result_builder.write(format(subheader_length, "04d"))
        if subheader_length > 0:
            result_builder.write(
                format(
                    NITFDESAccessor.parse_field_value(des_element, "DESSHF", str),
                    f">{subheader_length}",
                )
            )

        return result_builder.getvalue()

    @staticmethod
    def extract_desdata_xml(des_element: ET.Element) -> Optional[str]:
        """
        This function attempts to extract a block of XML from the field element named DESDATA. Versions of GDAL
        before 3.9 returned the XML data base64 encoded as a value attribute. Versions >=3.9 are automatically
        expanding the xml into the text area of an <xml_content> element.

        :param des_element: the root xml:DES metadata element
        :return: the xml content if it is found and can be extracted
        """
        desdata_element = des_element.find("./field[@name='DESDATA']")
        if desdata_element is None:
            return None

        value_attribute = desdata_element.get("value")
        if value_attribute:
            # This appears to be the encoding used by GDAL versions <3.9. Extract the
            # XML from the base64 encoded value attribute
            xml_bytes = base64.b64decode(value_attribute)
            return xml_bytes.decode("utf-8")

        xml_content_element = desdata_element.find("xml_content")
        if xml_content_element:
            # This appears to be a encoding used by GDAL >3.9. The XML is already parsed
            # and available as the content of this element. See: https://github.com/OSGeo/gdal/pull/8953
            return ET.tostring(xml_content_element[0], "unicode")

        # Unable to parse the XML from the data segment. This sometimes happens if GDAL
        # changes the representation of this information in their APIs
        return None

    @staticmethod
    def parse_field_value(des_element: ET.Element, field_name: str, type_conversion: Callable[[str], T]) -> T:
        """
        This is method will find a named "field" element in the children of a TRE Element and
        return the "value" attribute of that named field. A type conversion function can be provided to convert the
        attribute value to a specific Python type (e.g. int, float, or str)

        :param des_element: the root element to find the named field in
        :param field_name: the name of the field element
        :param type_conversion: the desired type of the output, must support construction from a string

        :return: the value converted to the requested type
        """
        field_element = des_element.find(f"./field[@name='{field_name}']")
        if field_element is None:
            raise ValueError(f"Unable to find TRE field named {field_name}")
        str_value = field_element.get("value")
        if str_value is None:
            raise ValueError(f"Field {field_name} does not have a value attribute.")

        return type_conversion(str_value)



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

# Telling flake8 to not flag errors in this file. It is normal that these classes are imported but not used in an
# __init__.py file.
# flake8: noqa
"""
The gdal package contains utilities that assist with loading imagery and metadata using the OSGeo GDAL library.

Loading Imagery and Sensor Models with OSML
*******************************************

OSML provides utilities to load a dataset and automatically construct an appropriate sensor model from metadata
available in the image. Metadata handled by GDAL (e.g. GeoTIFF tags or NITF segment metadata and TREs) is available
through the dataset accessors.

.. code-block:: python
    :caption: Example of loading a dataset and sensor model using OSML

    from aws.osml.gdal import load_gdal_dataset

    # Load the image and create a sensor model
    dataset, sensor_model = load_gdal_dataset("./imagery/sample.nitf")
    width = dataset.RasterXSize
    height = dataset.RasterYSize

    print(f"Loaded image with dimensions: ({height}, {width}) (rows, cols)")
    print(f"Using Sensor Model Implementation: {type(sensor_model).__name__}")
    print(dataset.GetMetadata())


Access to NITF Data Extension Segments
**************************************

SICD and SIDD imagery contains additional metadata in a XML Data Extension Segment that is not currently processed
by GDAL. This information can be accessed with the help of the NITFDESAccessor.

.. code-block:: python
    :caption: Example of loading a dataset and sensor model using OSML

    import base64
    import xml.dom.minidom
    from aws.osml.gdal import load_gdal_dataset, NITFDESAccessor

    dataset, sensor_model = load_gdal_dataset("./sample-sicd.nitf")

    des_accessor = NITFDESAccessor(dataset.GetMetadata("xml:DES"))
    xml_data_content_segments = des_accessor.get_segments_by_name("XML_DATA_CONTENT")
    if xml_data_content_segments is not None:
        for xml_data_segment in xml_data_content_segments:
            xml_str = des_accessor.extract_desdata_xml(xml_data_segment)
            if "SICD" in xml_str:
                temp = xml.dom.minidom.parseString(xml_str)
                new_xml = temp.toprettyxml()
                print(new_xml)
                break

-------------------------

APIs
****
"""

from .gdal_config import GDALConfigEnv, set_gdal_default_configuration
from .gdal_dem_tile_factory import GDALDigitalElevationModelTileFactory
from .gdal_utils import get_image_extension, get_type_and_scales, load_gdal_dataset
from .nitf_des_accessor import NITFDESAccessor
from .sensor_model_factory import ChippedImageInfoFacade, SensorModelFactory, SensorModelTypes
from .typing import GDALCompressionOptions, GDALImageFormats, RangeAdjustmentType

__all__ = [
    "set_gdal_default_configuration",
    "load_gdal_dataset",
    "get_image_extension",
    "get_type_and_scales",
    "GDALCompressionOptions",
    "GDALConfigEnv",
    "GDALDigitalElevationModelTileFactory",
    "GDALImageFormats",
    "RangeAdjustmentType",
    "NITFDESAccessor",
    "ChippedImageInfoFacade",
    "SensorModelFactory",
    "SensorModelTypes",
]



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

"""This file was generated by xsdata, v23.8, on 2023-10-05 09:59:45

Generator: DataclassGenerator
See: https://xsdata.readthedocs.io/
"""
from .cvenum_ism25_x import CVEnumISM25X
from .cvenum_ismclassification_all import CVEnumISMClassificationAll
from .cvenum_ismcomplies_with import CVEnumISMCompliesWithValues
from .cvenum_ismdissem import CVEnumISMDissemValuesvalue
from .cvenum_ismfgiopen import CVEnumISMFGIOpenValues
from .cvenum_ismfgiprotected import CVEnumISMFGIProtectedValues
from .cvenum_ismnon_ic import CVEnumISMNonICValues
from .cvenum_ismnon_uscontrols import CVEnumISMNonUSControlsValues
from .cvenum_ismowner_producer import CVEnumISMOwnerProducerValues
from .cvenum_ismrel_to import CVEnumISMRelToValues
from .cvenum_ismscicontrols import CVEnumISMSCIControlsValuesvalue
from .cvenum_ismsource_marked import CVEnumISMSourceMarked

__all__ = [
    "CVEnumISM25X",
    "CVEnumISMClassificationAll",
    "CVEnumISMCompliesWithValues",
    "CVEnumISMDissemValuesvalue",
    "CVEnumISMFGIOpenValues",
    "CVEnumISMFGIProtectedValues",
    "CVEnumISMNonICValues",
    "CVEnumISMNonUSControlsValues",
    "CVEnumISMOwnerProducerValues",
    "CVEnumISMRelToValues",
    "CVEnumISMSCIControlsValuesvalue",
    "CVEnumISMSourceMarked",
]



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

from math import degrees, floor, radians
from typing import Optional

from .coordinates import GeodeticWorldCoordinate
from .digital_elevation_model import DigitalElevationModelTileSet


class GenericDEMTileSet(DigitalElevationModelTileSet):
    """
    A generalizable tile set with a naming convention that can be described as a format string.
    """

    def __init__(
        self,
        format_spec: str = "%od%oh/%ld%lh.dt2",
        min_latitude_degrees: float = -90.0,
        max_latitude_degrees: float = 90.0,
        min_longitude_degrees: float = -180.0,
        max_longitude_degrees: float = 180.0,
    ) -> None:
        """
        Construct a tile set from a limited collection of configurable parameters. This implementation uses the
        custom formatting directives %od (longitude degrees), %oh (longitude hemisphere), %ld (latitude degrees),
        and %lh (latitude hemisphere) to allow users to create tile IDs that match a variety of situations.
        For example the default format_spec of '%od%oh/%ld%lh.dt2' will generate tile ids like: '115e/45s.dt2'
        which would match some common 1-degree cell based DEM file hierarchies.

        :param format_spec: the format specification for the GeodeteticWorldCoordinate


        :return: None
        """
        super().__init__()
        self.format_string = format_spec
        self.min_latitude_degrees = min_latitude_degrees
        self.max_latitude_degrees = max_latitude_degrees
        self.min_longitude_degrees = min_longitude_degrees
        self.max_longitude_degrees = max_longitude_degrees

    def find_tile_id(self, geodetic_world_coordinate: GeodeticWorldCoordinate) -> Optional[str]:
        """
        This method creates tile IDs that based on the format string provided.

        :param geodetic_world_coordinate: the world coordinate of interest

        :return: the tile path or None if the DEM does not have coverage for this location
        """
        longitude_degrees = floor(degrees(geodetic_world_coordinate.longitude))
        latitude_degrees = floor(degrees(geodetic_world_coordinate.latitude))

        # The SRTM mission only covers latitudes N59 through S56 so if the requested location is outside those
        # ranges we know there is no file available for it.
        if (
            latitude_degrees > self.max_latitude_degrees
            or latitude_degrees < self.min_latitude_degrees
            or longitude_degrees > self.max_longitude_degrees
            or longitude_degrees < self.min_longitude_degrees
        ):
            return None

        ul_coordinate = GeodeticWorldCoordinate([radians(longitude_degrees), radians(latitude_degrees), 0.0])
        return f"{ul_coordinate:{self.format_string}}"



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

from __future__ import annotations

import numpy as np
import numpy.typing as npt


class ProjectiveTransform:
    """
    This is a simple standalone projective transform class with an implementation that only depends on NumPy. There
    are equivalent classes in Open CV and Scikit Imaging but this class can be used when we don't want to include those
    dependencies.
    """

    def __init__(self, matrix_parameters: npt.ArrayLike) -> None:
        """
        Construct a projective transform from the given matrix parameters. Normally this constructor is not called
        directly. See the ProjectiveTransform.estimate() method instead.

        :param matrix_parameters: the matrix parameters for this transformation

        :return: None
        """
        self.matrix_parameters = matrix_parameters

    def forward(self, src_coords: npt.ArrayLike):
        """
        Compute the forward src -> dst transformation for an array of source coordinates [[x, y], ...]

        :param src_coords: the source coordinates

        :return: the array of transformed coordinates [[x',y'], ...]
        """
        a0, a1, a2, b0, b1, b2, c0, c1 = self.matrix_parameters
        x = src_coords[:, 0]
        y = src_coords[:, 1]
        out = np.zeros(src_coords.shape)
        out[:, 0] = (a0 + a1 * x + a2 * y) / (1 + c0 * x + c1 * y)
        out[:, 1] = (b0 + b1 * x + b2 * y) / (1 + c0 * x + c1 * y)
        return out

    def inverse(self, dst_coords: npt.ArrayLike):
        """
        Compute the inverse, dst -> src, transformation for an array of destination coordinates [[x', y'], ...]

        :param dst_coords: the destination coordinates

        :return: the array of transformed coordinates [[x,y], ...]
        """
        a0, a1, a2, b0, b1, b2, c0, c1 = self.matrix_parameters
        x = dst_coords[:, 0]
        y = dst_coords[:, 1]
        out = np.zeros(dst_coords.shape)
        out[:, 0] = (a2 * b0 - a0 * b2 + (b2 - b0 * c1) * x + (a0 * c1 - a2) * y) / (
            a1 * b2 - a2 * b1 + (b1 * c1 - b2 * c0) * x + (a2 * c0 - a1 * c1) * y
        )
        out[:, 1] = (a0 * b1 - a1 * b0 + (b0 * c0 - b1) * x + (a1 - a0 * c0) * y) / (
            a1 * b2 - a2 * b1 + (b1 * c1 - b2 * c0) * x + (a2 * c0 - a1 * c1) * y
        )
        return out

    @classmethod
    def estimate(cls, src: npt.ArrayLike, dst: npt.ArrayLike) -> ProjectiveTransform:
        """
        This method takes a list of source and destination points [x, y] and then applies the least squares fit
        to estimate the projective transform matrix relating them. It then creates the ProjectiveTransform
        object using that matrices. Each list needs to contain at least 4 points.

        :param src: the source points
        :param dst: the destination points

        :return: the projective transform
        """
        xs = src[:, 0]
        ys = src[:, 1]
        num_points = src.shape[0]

        # The Coefficient Matrix
        a = np.zeros((num_points * 2, 8))
        a[:num_points, 0] = 1
        a[:num_points, 1] = xs
        a[:num_points, 2] = ys
        a[num_points:, 3] = 1
        a[num_points:, 4] = xs
        a[num_points:, 5] = ys
        a[:num_points, 6] = -dst[:, 0] * xs
        a[:num_points, 7] = -dst[:, 0] * ys
        a[num_points:, 6] = -dst[:, 1] * xs
        a[num_points:, 7] = -dst[:, 1] * ys

        # The dependent variable values
        b = np.zeros((num_points * 2,))
        b[:num_points] = dst[:, 0]
        b[num_points:] = dst[:, 1]

        # See: https://numpy.org/doc/stable/reference/generated/numpy.linalg.lstsq.html
        matrix_parameters = np.linalg.lstsq(a, b, rcond=None)[0]

        return cls(matrix_parameters)



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import logging
from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt

from . import ElevationModel
from .coordinates import (
    GeodeticWorldCoordinate,
    ImageCoordinate,
    WorldCoordinate,
    geocentric_to_geodetic,
    geodetic_to_geocentric,
)
from .sensor_model import SensorModel

logger = logging.getLogger(__name__)


class Polynomial2D:
    """
    This class contains coefficients for a two-dimensional polynomial.
    """

    def __init__(self, coef: npt.ArrayLike):
        """
        Constructor that takes the coefficients of the polynomial. The coefficients should be ordered so that the
        coefficient of the term of multi-degree i,j is contained in coef[i,j].

        :param coef: array-like structure of coefficients
        """
        self.coef = np.array(coef)
        if len(self.coef.shape) != 2:
            raise ValueError(
                f"Coefficients for class Polynomial2D must be two-dimensional. "
                f"Received numpy.ndarray of shape {self.coef.shape}"
            )

    def __call__(self, x: Union[float, np.ndarray], y: Union[float, np.ndarray]) -> np.ndarray:
        """
        Invoke NumPy's polyval2d given the inputs and the coefficients of the polynomial.

        :param x: the first input parameter
        :param y: the second input parameter
        :return: the values of the 2-d polynomial at points formed with pairs of corresponding values from x and y.
        """
        return np.polynomial.polynomial.polyval2d(x, y, self.coef)


class PolynomialXYZ:
    """
    This class is an aggregation 3 one-dimensional polynomials all with the same input variable. The result of
    evaluating this class on the input variable is an [x, y, z] vector.
    """

    def __init__(
        self,
        x_polynomial: np.polynomial.Polynomial,
        y_polynomial: np.polynomial.Polynomial,
        z_polynomial: np.polynomial.Polynomial,
    ):
        """
        Constructor that accepts the 3 NumPy 1-d polynomials one for each component.

        :param x_polynomial: polynomial for the x component
        :param y_polynomial: polynomial for the y component
        :param z_polynomial: polynomial for the z component
        """
        self.x_polynomial = x_polynomial
        self.y_polynomial = y_polynomial
        self.z_polynomial = z_polynomial

    def __call__(self, t: float) -> np.ndarray:
        """
        Evaluate the x, y, and z polynomials at t and return the result as a vector.

        :param t: the value
        :return: the polynomial result
        """
        x = self.x_polynomial(t)
        y = self.y_polynomial(t)
        z = self.z_polynomial(t)

        return np.array([x, y, z], dtype=x.dtype)

    def deriv(self, m: int = 1):
        """
        Create a new PolynomialXYZ that is the derivative of the current PolynomialXYZ.

        :param m: find the derivative of order m
        :return: the new polynomial derivative
        """
        x_derivative = self.x_polynomial.deriv(m=m)
        y_derivative = self.y_polynomial.deriv(m=m)
        z_derivative = self.z_polynomial.deriv(m=m)

        return PolynomialXYZ(x_polynomial=x_derivative, y_polynomial=y_derivative, z_polynomial=z_derivative)


class SARImageCoordConverter:
    """
    This class contains image grid and image plane coordinate conversions for a provided set of SICD parameters. The
    equations are mostly defined in Section 2 of the SICD Standard Volume 3.
    """

    def __init__(
        self,
        scp_pixel: ImageCoordinate,
        scp_ecf: WorldCoordinate,
        u_row: np.ndarray,
        u_col: np.ndarray,
        row_ss: float,
        col_ss: float,
        first_pixel: ImageCoordinate = ImageCoordinate([0.0, 0.0]),
    ):
        """
        Construct the coordinate converter given parameters from the metadata. The names of these parameters have been
        chosen to align with the names in the specification.

        :param scp_pixel: location of the scene center point (SCP) in the global pixel grid
        :param scp_ecf: location of the scene center point (SCP) in earth centered fixed (ECF) coordinates
        :param u_row: unit vector in the increasing row direction in ECF coordinates.
        :param u_col: unit vector in the increasing column direction in ECF coordinates.
        :param row_ss: sample spacing in the row direction
        :param col_ss: sample spacing in the column direction
        :param first_pixel: location of the first row/column of the pixel array. For a full image array [0, 0]
        """
        self.scp_pixel = scp_pixel
        self.scp_ecf = scp_ecf
        self.row_ss = row_ss
        self.col_ss = col_ss
        self.u_row = u_row
        self.u_col = u_col
        self.first_pixel = first_pixel
        # Section 2.4 calculation of the image plane unit normal vector
        ipn = np.cross(self.u_row, self.u_col)
        self.uvect_ipn = ipn / np.linalg.norm(ipn)

        # Section 2.4 calculation of transform from ipp to xrow, ycol
        cos_theta = np.dot(self.u_row, self.u_col)
        sin_theta = np.sqrt(1 - cos_theta * cos_theta)
        ipp_transform = np.array([[1, -cos_theta], [-cos_theta, 1]], dtype="float64") / (sin_theta * sin_theta)
        row_col_transform = np.zeros((3, 2), dtype="float64")
        row_col_transform[:, 0] = self.u_row
        row_col_transform[:, 1] = self.u_col
        self.matrix_transform = np.dot(row_col_transform, ipp_transform)

    def rowcol_to_xrowycol(self, row_col: np.ndarray) -> np.ndarray:
        """
        This function converts the row and column indexes (row, col) in the global image grid to SCP centered
        image coordinates (xrow, ycol) using equations (2) (3) in Section 2.2 of the SICD Specification
        Volume 3.

        :param row_col: the [row, col] location as an array
        :return: the [xrow, ycol] location as an array
        """
        xrow_ycol = np.zeros(2, dtype="float64")
        xrow_ycol[0] = (row_col[0] - self.scp_pixel.r) * self.row_ss
        xrow_ycol[1] = (row_col[1] - self.scp_pixel.c) * self.col_ss
        return xrow_ycol

    def xrowycol_to_rowcol(self, xrow_ycol: np.ndarray) -> np.ndarray:
        """
        This function converts the SCP centered image coordinates (xrow, ycol) to row and column indexes (row, col)
        in the global image grid using equations (2) (3) in Section 2.2 of the SICD Specification Volume 3.

        :param xrow_ycol: the [xrow, ycol] location as an array
        :return: the [row, col] location as an array
        """
        row_col = np.zeros(2, dtype="float64")
        row_col[0] = xrow_ycol[0] / self.row_ss + self.scp_pixel.r
        row_col[1] = xrow_ycol[1] / self.col_ss + self.scp_pixel.c
        return row_col

    def xrowycol_to_ipp(self, xrow_ycol: np.ndarray) -> np.ndarray:
        """
        This function converts SCP centered image coordinates (xrow, ycol) to a ECF coordinate, image plane point (IPP),
        on the image plane using equations in Section 2.4 of the SICD Specification Volume 3.

        :param xrow_ycol: the [xrow, ycol] location as an array
        :return: the image plane point [x, y, z] ECF location on the image plane
        """
        delta_ipp = xrow_ycol[0] * self.u_row + xrow_ycol[1] * self.u_col
        return self.scp_ecf.coordinate + delta_ipp

    def ipp_to_xrowycol(self, ipp: np.ndarray) -> np.ndarray:
        """
        This function converts an ECF location on the image plane into SCP centered image coordinates (xrow, ycol)
        using equations in Section 2.4 of the SICD Specification volume 3.

        :param ipp: the image plane point [x, y, z] ECF location on the image plane
        :return: the [xrow, ycol] location as an array
        """
        delta_ipp = ipp - self.scp_ecf.coordinate
        xrow_ycol = np.dot(delta_ipp, self.matrix_transform)
        return xrow_ycol


class COAProjectionSet:
    """
    This is an abstract base class for R/Rdot projection contour computations described in Section 4 of the SICD
    Standard Volume 3.
    """

    def __init__(
        self,
        coa_time_poly: Polynomial2D,
        arp_poly: PolynomialXYZ,
        delta_arp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        delta_varp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        range_bias: float = 0.0,
    ):
        """
        Constructor with parameters supporting the calculations common to all R/Rdot projections (i.e. the
        calculations that do not depend on grid type and image formation algorithm).

        :param coa_time_poly: Center Of Aperture (COA) time polynomial.
        :param arp_poly: Aperture Reference Point (ARP) position polynomial coefficients.
        :param delta_arp: the ARP position offset
        :param delta_varp: the ARP velocity offset
        :param range_bias: the range bias offset
        """
        self.coa_time_poly = coa_time_poly
        self.arp_poly = arp_poly
        self.varp_poly = self.arp_poly.deriv(m=1)

        self.delta_arp = delta_arp
        self.delta_varp = delta_varp
        self.range_bias = float(range_bias)

    def precise_rrdot_computation(
        self, xrow_ycol: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        This executes the precise image pixel grid location to R/Rdot projection. This function invokes
        the _grid_specific_projection() function implemented by subclasses which should handle the portions
        of the calculation that are dependent on the image grid and image formation algorithm.

        :param xrow_ycol: the [xrow, ycol] location as an array
        :return: the COA projection set { Rcoa, Rdotcoa, tcoa, arpcoa, varpcoa }
        """
        # These are the common calculations for image COA time (coa_time), COA ARP position and velocity
        # (arp_position and arp_velocity) as described in Section 2 of the SICD specification Volume 3.
        coa_time = self.coa_time_poly(xrow_ycol[0], xrow_ycol[1])
        arp_position = self.arp_poly(coa_time)
        arp_velocity = self.varp_poly(coa_time)

        # These are the image grid and image formation algorithm dependent calculations for the precise
        # computation of the R/Rdot contour. Each subclass should implement an approach as described in
        # sections 4.1 through 4.6 of the SICD specification Volume 3.
        r_tgt_coa, r_dot_tgt_coa = self._grid_specific_projection(xrow_ycol, coa_time, arp_position, arp_velocity)

        # If provided the Adjustable Parameter Offsets are incorporated into the computation from
        # image pixel location to COA projection parameters. See Section 8.1 of the SICD specification Volume 3.
        # TODO: Check this. This is the same approach as SarPy but I'm not 100% sure it is correct
        arp_position += self.delta_arp
        arp_velocity += self.delta_varp
        r_tgt_coa += self.range_bias

        return r_tgt_coa, r_dot_tgt_coa, coa_time, arp_position, arp_velocity

    @abstractmethod
    def _grid_specific_projection(self, xrow_ycol, coa_time, arp_position, arp_velocity) -> Tuple[np.ndarray, np.ndarray]:
        """
        The precise computation of the R/Rdot contour is dependent upon the image grid type and the image
        formation algorithm that produced the image but the computation of the COA time, ARP position, and
        velocity is the same for all image products.

        This abstract method should be overriden by subclasses to perform the R/Rdot calculations for a
        specific image grid and formation algorithm.

        :param xrow_ycol: the [xrow, ycol] location as an array
        :param coa_time: Center Of Aperture (COA) time
        :param arp_position: Aperture Reference Point (ARP) position
        :param arp_velocity: Aperture Reference Point (ARP) velocity
        :return: the tuple containing range and range rate relative to the ARP at COA time
        """


class PFAProjectionSet(COAProjectionSet):
    """
    This Center Of Aperture (COA) Projection set is to be used with a range azimuth image grid (RGAZIM) and polar
    formatted (PFA) phase history data. See section 4.1 of the SICD Specification Volume 3.
    """

    def __init__(
        self,
        scp_ecf: WorldCoordinate,
        polar_ang_poly,
        spatial_freq_sf_poly,
        coa_time_poly: Polynomial2D,
        arp_poly: PolynomialXYZ,
        delta_arp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        delta_varp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        range_bias: float = 0.0,
    ):
        """
        Constructor for this projection set.

        :param scp_ecf: Scene Center Point position in ECF coordinates
        :param polar_ang_poly: Polar Angle polynomial coefficients
        :param spatial_freq_sf_poly: Spatial Frequency Scale Factor polynomial coefficients
        :param coa_time_poly: Center Of Aperture (COA) time polynomial
        :param arp_poly: Aperture Reference Point (ARP) position polynomial coefficients
        :param delta_arp: the ARP position offset
        :param delta_varp: the ARP velocity offset
        :param range_bias: the range bias offset
        """
        super().__init__(coa_time_poly, arp_poly, delta_arp, delta_varp, range_bias)
        self.scp_ecf = scp_ecf
        self.polar_ang_poly = polar_ang_poly
        self.spatial_freq_sf_poly = spatial_freq_sf_poly
        self.polar_ang_poly_der = polar_ang_poly.deriv(m=1)
        self.spatial_freq_sf_poly_der = spatial_freq_sf_poly.deriv(m=1)

    def _grid_specific_projection(self, xrow_ycol, coa_time, arp_position, arp_velocity) -> Tuple[np.ndarray, np.ndarray]:
        """
        These are the calculations for the precise computation of the R/Rdot contour unique to these grid and
        image formation algorithm types. See SICD Volume 3 Section 4.1

        :param xrow_ycol: the [xrow, ycol] location as an array
        :param coa_time: Center Of Aperture (COA) time
        :param arp_position: Aperture Reference Point (ARP) position
        :param arp_velocity: Aperture Reference Point (ARP) velocity
        :return: the tuple containing range and range rate relative to the ARP at COA time
        """
        # For the RGAZIM grid, the image coordinates are range and azimuth. The row coordinate is the range
        # coordinate, xrow = rg. The column coordinate is the azimuth coordinate, ycol = az.
        rg = xrow_ycol[0]
        az = xrow_ycol[1]

        # (2) Compute the range and range rate to the SCP at the pixel COA time
        arp_minus_scp = arp_position - self.scp_ecf.coordinate
        range_to_scp = np.linalg.norm(arp_minus_scp, axis=-1)
        rdot_to_scp = np.sum(arp_velocity * arp_minus_scp, axis=-1) / range_to_scp

        # (3) Compute the polar angle (theta) and its derivative with respect to time (d_theta_d_time)
        # at the pixel COA time.
        theta = self.polar_ang_poly(coa_time)
        d_theta_d_time = self.polar_ang_poly_der(coa_time)

        # (4) Compute the polar aperture scale factor (KSF) and its derivative with respect to polar angle
        # (d_ksf_d_theta) at the pixel COA time
        ksf = self.spatial_freq_sf_poly(theta)
        d_ksf_d_theta = self.spatial_freq_sf_poly_der(theta)

        # (5) Compute the spatial frequency domain phase slopes in the radial (ka) and cross radial
        # (kc) directions (d_phi_d_ka and d_phi_d_kc) for the radial direction at theta. Note: The sign COA
        # parameter (SGN) for the phase may be ignored as it is cancelled in a subsequent computation.
        d_phi_d_ka = rg * np.cos(theta) + az * np.sin(theta)
        d_phi_d_kc = -rg * np.sin(theta) + az * np.cos(theta)

        # (6) Compute range relative to the SCP (delta_range) at the COA.
        delta_range = ksf * d_phi_d_ka

        # (7) Compute the derivative of the range relative to the SCP with respect to polar angle
        # (d_delta_range_d_theta) at the COA. Scale by the derivative of the polar angle with respect
        # to time to yield the derivative with respect to time (delta_r_dot_tgt_coa).
        d_delta_range_d_theta = d_ksf_d_theta * d_phi_d_ka + ksf * d_phi_d_kc
        delta_r_dot_tgt_coa = d_delta_range_d_theta * d_theta_d_time

        # (8) Compute the range and range rate relative to the ARP at COA ( r_tgt_coa and rdot_tgt_coa).
        # The projection to three-dimensional scene point for grid location (rgTGT, azTGT) is along this
        # R/Rdot contour.
        r_tgt_coa = range_to_scp + delta_range
        rdot_tgt_coa = rdot_to_scp + delta_r_dot_tgt_coa

        return r_tgt_coa, rdot_tgt_coa


class RGAZCOMPProjectionSet(COAProjectionSet):
    def __init__(
        self,
        scp_ecf: WorldCoordinate,
        az_scale_factor: float,
        coa_time_poly: Polynomial2D,
        arp_poly: PolynomialXYZ,
        delta_arp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        delta_varp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        range_bias: float = 0.0,
    ):
        """
        Constructor for this projection set.

        :param scp_ecf: Scene Center Point position in ECF coordinates
        :param az_scale_factor: Scale factor that converts azimuth coordinate to an increment in cosine of the DCA at COA
        :param coa_time_poly: Center Of Aperture (COA) time polynomial
        :param arp_poly: Aperture Reference Point (ARP) position polynomial coefficients
        :param delta_arp: the ARP position offset
        :param delta_varp: the ARP velocity offset
        :param range_bias: the range bias offset
        """
        super().__init__(coa_time_poly, arp_poly, delta_arp, delta_varp, range_bias)
        self.scp_ecf = scp_ecf
        self.az_scale_factor = az_scale_factor

    def _grid_specific_projection(self, xrow_ycol, coa_time, arp_position, arp_velocity) -> Tuple[np.ndarray, np.ndarray]:
        """
        These are the calculations for the precise computation of the R/Rdot contour unique to these grid and
        image formation algorithm types. See SICD Volume 3 Section 4.2

        :param xrow_ycol: the [xrow, ycol] location as an array
        :param coa_time: Center Of Aperture (COA) time
        :param arp_position: Aperture Reference Point (ARP) position
        :param arp_velocity: Aperture Reference Point (ARP) velocity
        :return: the tuple containing range and range rate relative to the ARP at COA time
        """
        # For the RGAZIM grid, the image coordinates are range and azimuth. The row coordinate is the range
        # coordinate, xrow = rg. The column coordinate is the azimuth coordinate, ycol = az.
        rg = xrow_ycol[0]
        az = xrow_ycol[1]

        # (2) Compute the range and range rate to the SCP at COA.
        arp_minus_scp = arp_position - self.scp_ecf.coordinate
        range_to_scp = np.linalg.norm(arp_minus_scp, axis=-1)
        rdot_to_scp = np.sum(arp_velocity * arp_minus_scp, axis=-1) / range_to_scp

        # (3) Compute the increment in cosine of the DCA at COA of the target (delta_cos_dca_tgt_coa) by
        # scaling the azimuth coordinate by the azimuth to DCA scale factor. Compute the increment
        # in range rate (delta_rdot_tgt_coa) by scaling by the magnitude of the velocity vector at COA.
        delta_cos_dca_tgt_coa = self.az_scale_factor * az
        delta_r_dot_tgt_coa = -np.linalg.norm(arp_velocity, axis=-1) * delta_cos_dca_tgt_coa

        # (4) Compute the range and range rate to the target at COA as follows.
        r_tgt_coa = range_to_scp + rg
        rdot_tgt_coa = rdot_to_scp + delta_r_dot_tgt_coa

        return r_tgt_coa, rdot_tgt_coa


class INCAProjectionSet(COAProjectionSet):
    def __init__(
        self,
        r_ca_scp: float,
        inca_time_coa_poly: np.polynomial.Polynomial,
        drate_sf_poly: Polynomial2D,
        coa_time_poly: Polynomial2D,
        arp_poly: PolynomialXYZ,
        delta_arp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        delta_varp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        range_bias: float = 0.0,
    ):
        """
        Constructor for this projection set.

        :param r_ca_scp: Range at Closest Approach for the SCP (m)
        :param inca_time_coa_poly: Time of Closest Approach polynomial coefficients
        :param drate_sf_poly: Doppler Rate Scale Factor polynomial coefficients
        :param coa_time_poly: Center Of Aperture (COA) time polynomial
        :param arp_poly: Aperture Reference Point (ARP) position polynomial coefficients
        :param delta_arp: the ARP position offset
        :param delta_varp: the ARP velocity offset
        :param range_bias: the range bias offset
        """
        super().__init__(coa_time_poly, arp_poly, delta_arp, delta_varp, range_bias)
        self.r_ca_scp = r_ca_scp
        self.inca_time_coa_poly = inca_time_coa_poly
        self.drate_sf_poly = drate_sf_poly

    def _grid_specific_projection(self, xrow_ycol, coa_time, arp_position, arp_velocity) -> Tuple[np.ndarray, np.ndarray]:
        """
        These are the calculations for the precise computation of the R/Rdot contour unique to these grid and
        image formation algorithm types. See SICD Volume 3 Section 4.3

        :param xrow_ycol: the [xrow, ycol] location as an array
        :param coa_time: Center Of Aperture (COA) time
        :param arp_position: Aperture Reference Point (ARP) position
        :param arp_velocity: Aperture Reference Point (ARP) velocity
        :return: the tuple containing range and range rate relative to the ARP at COA time
        """
        # For the RGZERO grid, the image coordinates are range and azimuth. The row coordinate is the range
        # coordinate, xrow = rg. The column coordinates is the azimuth coordinate, ycol = az.
        rg = xrow_ycol[0]
        az = xrow_ycol[1]

        # (2) Compute the range at closest approach and the time of closest approach for the image
        # grid location. The range at closest approach, R TGT , is computed from the range coordinate.
        # The time of closest approach, tTGT , is computed from the azimuth coordinate. CA
        range_ca_tgt = self.r_ca_scp + rg
        time_ca_tgt = self.inca_time_coa_poly(az)

        # (2 repeated in v1.3.0 of the spec) Compute the ARP velocity at the time of closest approach
        # and the magnitude of the vector.
        arp_velocity_ca_tgt = self.varp_poly(time_ca_tgt)
        mag_arp_velocity_ca_tgt = np.linalg.norm(arp_velocity_ca_tgt)

        # (3) Compute the Doppler Rate Scale Factor (drsf_tgt) for image grid location (rg, az).
        drsf_tgt = self.drate_sf_poly(rg, az)

        # (4) Compute the time difference between the COA time and the CA time (delta_coa_tgt).
        delta_coa_tgt = coa_time - time_ca_tgt

        # (5) Compute the range and range rate relative to the ARP at COA ( RTGT and RdotTGT ).
        r_tgt_coa = np.sqrt(range_ca_tgt**2 + drsf_tgt * mag_arp_velocity_ca_tgt**2 * delta_coa_tgt**2)
        r_dot_tgt_coa = (drsf_tgt / r_tgt_coa) * mag_arp_velocity_ca_tgt**2 * delta_coa_tgt

        return r_tgt_coa, r_dot_tgt_coa


class PlaneProjectionSet(COAProjectionSet):
    def __init__(
        self,
        scp_ecf: WorldCoordinate,
        image_plane_urow: np.ndarray,
        image_plane_ucol: np.ndarray,
        coa_time_poly: Polynomial2D,
        arp_poly: PolynomialXYZ,
        delta_arp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        delta_varp: np.ndarray = np.array([0.0, 0.0, 0.0], dtype="float64"),
        range_bias: float = 0.0,
    ):
        """
        Constructor for this projection set.

        :param scp_ecf: Scene Center Point position in ECF coordinates
        :param image_plane_urow: Unit vector in the increasing row direction in ECF coordinates (uRow)
        :param image_plane_ucol: Unit vector in the increasing column direction in ECF coordinates (uCol)
        :param coa_time_poly: Center Of Aperture (COA) time polynomial
        :param arp_poly: Aperture Reference Point (ARP) position polynomial coefficients
        :param delta_arp: the ARP position offset
        :param delta_varp: the ARP velocity offset
        :param range_bias: the range bias offset
        """
        super().__init__(coa_time_poly, arp_poly, delta_arp, delta_varp, range_bias)
        self.scp_ecf = scp_ecf
        self.image_plane_urow = image_plane_urow
        self.image_plane_ucol = image_plane_ucol

    def _grid_specific_projection(self, xrow_ycol, coa_time, arp_position, arp_velocity) -> Tuple[np.ndarray, np.ndarray]:
        """
        These are the calculations for the precise computation of the R/Rdot contour unique to these grid and
        image formation algorithm types. See SICD Volume 3 Sections 4.4, 4.5, and 4.6.

        Note that the calculations in sections 4.4, 4.5, and 4.6 are the same with the only difference being the
        interpretation of the xrow and ycol gird positions. To share this one implementation for all three grid
        planes assume: xrow = xrg = xct and ycol = ycr = yat

        :param xrow_ycol: the [xrow, ycol] location as an array
        :param coa_time: Center Of Aperture (COA) time
        :param arp_position: Aperture Reference Point (ARP) position
        :param arp_velocity: Aperture Reference Point (ARP) velocity
        :return: the tuple containing range and range rate relative to the ARP at COA time
        """

        # xrow = xrg = xct and ycol = ycr = yat

        # (2) The samples of an XRGYCR, XCTYAT, or PLANE grid are uniformly spaced locations in the image plane
        # formed by the SCP, and image plane vectors uRow and uCol. Vectors uRow and uCol are orthogonal. Compute
        # the point the image plane point for image grid location (xrgTGT, ycrTGT).
        image_plane_point = (
            self.scp_ecf.coordinate + xrow_ycol[0] * self.image_plane_urow + xrow_ycol[1] * self.image_plane_ucol
        )

        # (3) Compute the range and range rate relative to the ARP at COA (r_tgt_coa and rdot_tgt_coa) for image plane
        # point (image_plane_point).
        arp_minus_ipp = arp_position - image_plane_point
        r_tgt_coa = np.linalg.norm(arp_minus_ipp, axis=-1)
        rdot_tgt_coa = np.sum(arp_velocity * arp_minus_ipp, axis=-1) / r_tgt_coa

        return r_tgt_coa, rdot_tgt_coa


class RRDotSurfaceProjection:
    """
    This is the base class for calculations that project the R/RDot contour onto a surface model. The SICD specification
    defines a way to do this for planes, a surface of constant height above an ellipsoid, or a digital elevation model.
    """

    @abstractmethod
    def rrdot_to_ground(self, r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity) -> np.ndarray:
        """
        Subclasses should implement this method to compute the R/RDot Contour Ground Plane intersection with a
        specific surface type (e.g. planar, HAE, DEM)

        :param r_tgt_coa: target COA range
        :param r_dot_tgt_coa: target COA range rate
        :param arp_position: ARP position
        :param arp_velocity: ARP velocity
        :return: the intersection between the R/Rdot Contour and the ground plane
        """


class GroundPlaneRRDotSurfaceProjection(RRDotSurfaceProjection):
    """
    This class implements the Precise R/RDot Ground Plane Projection described in Section 5 of the SICD Specification
    Volume 3 (v1.3.0).
    """

    class GroundPlaneNormalType(Enum):
        SPHERICAL_EARTH = "SPHERICAL_EARTH"
        GEODETIC_EARTH = "GEODETIC_EARTH"

    def __init__(
        self,
        ref_ecf: WorldCoordinate,
        gpn: Optional[np.ndarray],
        gpn_type: GroundPlaneNormalType = GroundPlaneNormalType.GEODETIC_EARTH,
    ):
        """
        The ground plane is defined by a reference point in the plane (ref_ect) and the vector normal to the plane
        (gpn). The reference point and plane orientation may be based upon specific terrain height and slope
        information for the imaged area. When only a reference point is specified, a ground plane normal may be
        derived assuming the plane is tangent to a spherical earth model or a surface of constant geodetic height
        above the WGS-84 reference ellipsoid passing through (ref_ect).

        :param ref_ecf: reference point in the plane, GREF in the specification
        :param gpn: optional vector normal to the ground plane; if missing it will be computed using gpn_type
        :param gpn_type: method to derive the ground plan normal
        """
        self.ref_ecf = ref_ecf

        if gpn is not None:
            self.u_gpn = gpn / np.linalg.norm(gpn)
        elif gpn_type == self.GroundPlaneNormalType.SPHERICAL_EARTH:
            self.u_gpn = ref_ecf.coordinate / np.linalg.norm(ref_ecf.coordinate)
        elif gpn_type == self.GroundPlaneNormalType.GEODETIC_EARTH:
            ref_lle = geocentric_to_geodetic(ref_ecf)
            self.u_gpn = np.array(
                [
                    np.cos(ref_lle.latitude) * np.cos(ref_lle.longitude),
                    np.cos(ref_lle.latitude) * np.sin(ref_lle.longitude),
                    np.sin(ref_lle.latitude),
                ]
            )
        else:
            raise ValueError(f"Provided gpn_type, {gpn_type}, is invalid.")

    def rrdot_to_ground(self, r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity) -> np.ndarray:
        """
        This method implements the R/RDot Contour Ground Plane Intersection described in section 5.2

        :param r_tgt_coa: target COA range
        :param r_dot_tgt_coa: target COA range rate
        :param arp_position: ARP position
        :param arp_velocity: ARP velocity
        :return: the intersection between the R/Rdot Contour and the ground plane
        """
        # (1) Compute the unit vector in the +Z direction (normal to the ground plane).
        uvect_z = self.u_gpn / np.linalg.norm(self.u_gpn)

        # (2) Compute the ARP distance from the plane (arp_z). Also compute the ARP ground plane nadir (agpn).
        arp_z = np.sum((arp_position - self.ref_ecf.coordinate) * uvect_z, axis=-1)
        if arp_z > r_tgt_coa:
            raise ValueError("No solution exists. Distance between ARP and the plane is greater than range.")

        agpn = arp_position - np.outer(arp_z, uvect_z)

        # (3) Compute the ground plane distance (gp_distance) from the ARP nadir to the circle of constant range. Also
        # compute the sine and cosine of the grazing angle (sin_graz and cos_graz).
        gp_distance = np.sqrt(r_tgt_coa * r_tgt_coa - arp_z * arp_z)
        sin_graz = arp_z / r_tgt_coa
        cos_graz = gp_distance / r_tgt_coa

        # (4) Compute velocity components normal to the ground plane (v_z) and parallel to the ground plane (v_x).
        v_z = np.dot(arp_velocity, uvect_z)
        v_mag = np.linalg.norm(arp_velocity, axis=-1)
        v_x = np.sqrt(v_mag * v_mag - v_z * v_z)

        # (5) Orient the +X direction in the ground plane such that the v_x > 0. Compute unit vectors uvect_x
        # and uvect_y.
        uvect_x = (arp_velocity - (v_z * uvect_z)) / v_x
        uvect_y = np.cross(uvect_z, uvect_x)

        # (6) Compute the cosine of the azimuth angle to the ground plane point.
        cos_az = (-r_dot_tgt_coa + v_z * sin_graz) / (v_x * cos_graz)
        if np.abs(cos_az) > 1:
            raise ValueError("No solution exists. cos_az < -1 or cos_az > 1.")

        # (7) Compute the sine of the azimuth angle. Use parameter LOOK to establish the correct sign corresponding
        # to the correct Side of Track.
        look = np.sign(np.dot(np.cross(arp_position - self.ref_ecf.coordinate, arp_velocity), uvect_z))
        sin_az = look * np.sqrt(1 - cos_az * cos_az)

        # (8) Compute GPPTGT at distance G from the AGPN and at the correct azimuth angle.
        return agpn + uvect_x * (gp_distance * cos_az) + uvect_y * (gp_distance * sin_az)


class HAERRDotSurfaceProjection(RRDotSurfaceProjection):
    """
    This class implements the Precise R/RDot Height Above Ellipsioid (HAE) Projection described in Section 9 of the
    SICD Specification Volume 3 (v1.3.0).
    """

    def __init__(
        self,
        scp_ecf: WorldCoordinate,
        side_of_track: str,
        hae: float,
        height_threshold: float = 1.0,
        max_number_iterations: int = 3,
    ):
        """
        Constructor for the projection that takes the image parameters and a height above theWGS-84 ellipsoid.
        The parameter defaults are the recommended values for the user selectable parameters in section 9.2.

        :param scp_ecf: Scene Center Point position in ECF coordinates
        :param side_of_track: side of track imaged
        :param hae: the surface height (m) above the WGS-84 reference ellipsoid
        :param height_threshold: the height threshold for convergence of iterative projection sequence
        :param max_number_iterations: maximum number of iterations allowed
        """
        self.scp_ecf = scp_ecf
        self.scp_lle = geocentric_to_geodetic(scp_ecf)
        self.side_of_track = side_of_track
        self.hae = hae

        # Integer based on side of track parameter
        self.look = 1.0
        if side_of_track == "R":
            self.look = -1.0

        self.delta_hae_max = height_threshold
        self.nlim = max_number_iterations

    def rrdot_to_ground(self, r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity) -> np.ndarray:
        """
        This method implements the R/RDot Contour Ground Plane Intersection described in section 9.2.

        The precise projection to a surface of constant height above the WGS-84 reference ellipsoid along an
        R/Rdot contour. The R/Rdot contour is relative to an ARP Center Of Aperture position and velocity. The
        algorithm computes the R/Rdot projection to one or more ground planes that are tangent to the constant
        height surface. Each ground plane projection point computed is slightly above the constant HAE surface.
        The final surface position is computed by projecting from the final ground plane projection point down
        to the HAE surface.

        :param r_tgt_coa: target COA range
        :param r_dot_tgt_coa: target COA range rate
        :param arp_position: ARP position
        :param arp_velocity: ARP velocity
        :return: the intersection between the R/Rdot Contour and the ground plane
        """
        # (1) Compute the geodetic ground plane normal at the SCP. Compute the parameters for the initial ground plane.
        # The reference point position is gref and the unit normal is u_gpn.
        u_gpn = np.array(
            [
                np.cos(self.scp_lle.latitude) * np.cos(self.scp_lle.longitude),
                np.cos(self.scp_lle.latitude) * np.sin(self.scp_lle.longitude),
                np.sin(self.scp_lle.latitude),
            ]
        )
        gref = self.scp_ecf.coordinate + (self.hae - self.scp_lle.z) * u_gpn

        cont = True
        n = 1
        while cont:
            # (2) Compute the precise projection along the R/Rdot contour to Ground Plane. The result is ground plane
            # point position gpp_ecf. Convert from ECF coordinates to geodetic coordinates (gpp_lle).
            gp_surface_projection = GroundPlaneRRDotSurfaceProjection(ref_ecf=WorldCoordinate(gref), gpn=u_gpn)
            gpp_ecf = gp_surface_projection.rrdot_to_ground(r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity)[0]
            gpp_lle = geocentric_to_geodetic(WorldCoordinate(coordinate=gpp_ecf))

            # (3) Compute the unit vector in the increasing height direction at point gpp_lle, (u_up). Also
            # compute the height difference at point gpp_lle relative to the desired surface height (delta_hae).
            u_up = np.array(
                [
                    np.cos(gpp_lle.latitude) * np.cos(gpp_lle.longitude),
                    np.cos(gpp_lle.latitude) * np.sin(gpp_lle.longitude),
                    np.sin(gpp_lle.latitude),
                ]
            )
            delta_hae = gpp_lle.elevation - self.hae

            # (4) Test to see if the point is sufficiently close the surface or if the maximum number of iterations
            # has been reached.  Otherwise, compute a new ground reference point (gref) and unit normal (u_up); repeat
            # Steps 2, 3 and 4.
            if delta_hae <= self.delta_hae_max or n >= self.nlim:
                cont = False
            else:
                gref = gpp_ecf - delta_hae * u_up
                u_gpn = u_up
                n += 1

        # (5) Compute the unit slant plane normal vector, u_spn, that is tangent to the R/Rdot contour at point gpp.
        # Unit vector u_spn points away from the center of the earth and in a direction of increasing HAE at gpp.
        spn = np.cross(self.look * arp_velocity, gpp_ecf - arp_position)
        u_spn = spn / np.linalg.norm(spn)

        # (6) Compute the straight line projection from point gpp_ecf along the slant plane normal to point slp.
        # Point slp is very close to the precise R/Rdot contour intersection with the constant height surface.
        # Convert the position of point slp from ECF coordinates to geodetic coordinates (slp_lle).
        sf = np.dot(u_up, u_spn)
        slp = gpp_ecf - (delta_hae / sf) * u_spn
        slp_lle = geocentric_to_geodetic(WorldCoordinate(slp))

        # (7) Assign surface point spp position by adjusting the HAE to be on the desired surface. Convert from
        # geodetic coordinates to ECF coordinates.
        spp_lle = GeodeticWorldCoordinate([slp_lle.longitude, slp_lle.latitude, self.hae])
        spp_ecf = geodetic_to_geocentric(spp_lle)

        return np.array([spp_ecf.coordinate])


class DEMRRDotSurfaceProjection(RRDotSurfaceProjection):
    """
    This class implements the Precise R/RDot Height Above a Digital Elevation Model (DEM) Projection described in
    Section 10 of the SICD Specification Volume 3 (v1.3.0).
    """

    def __init__(
        self,
        scp_ecf: WorldCoordinate,
        side_of_track: str,
        elevation_model: ElevationModel,
        max_adjacent_point_distance: float = 10.0,
        height_threshold: float = 0.001,
    ):
        """
        Constructor for the projection that takes the image parameters and a digital elevation model (DEM).
        The parameter defaults are the recommended values for the user selectable parameters in section 10.3.

        :param scp_ecf: Scene Center Point position in ECF coordinates
        :param side_of_track: side of track imaged
        :param elevation_model: the digital elevation model
        :param max_adjacent_point_distance: Maximum distance between adjacent points along the R/Rdot contour
        :param height_threshold: threshold for determining if a R/Rdot contour point is on the DEM surface (m)
        """
        self.scp_ecf = scp_ecf
        self.scp_lle = geocentric_to_geodetic(scp_ecf)
        self.side_of_track = side_of_track
        self.elevation_model = elevation_model

        # Integer based on Side of Track parameter.
        self.look = 1.0
        if side_of_track == "R":
            self.look = -1.0

        elevation_summary = elevation_model.describe_region(self.scp_lle)
        self.hae_min = elevation_summary.min_elevation
        self.hae_max = elevation_summary.max_elevation
        self.hae_max_surface_projection = HAERRDotSurfaceProjection(
            scp_ecf=scp_ecf, side_of_track=side_of_track, hae=self.hae_max
        )
        self.hae_min_surface_projection = HAERRDotSurfaceProjection(
            scp_ecf=scp_ecf, side_of_track=side_of_track, hae=self.hae_min
        )
        self.delta_dist_dem = 0.5 * elevation_summary.post_spacing

        self.delta_dist_rrc = max_adjacent_point_distance
        self.delta_hd_lim = height_threshold

    def rrdot_to_ground(self, r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity) -> np.ndarray:
        """
        This method implements the R/RDot Contour Ground Plane Intersection described in section 10.3

        The R/Rdot contour is relative to an ARP Center Of Aperture position and velocity. The earth surface is
        described by a Digital Elevation Model (DEM) that defines a unique surface height as a function of two
        horizontal coordinates. The projection computation may yield one or more surface points that lie along
        the R/Rdot contour.

        :param r_tgt_coa: target COA range
        :param r_dot_tgt_coa: target COA range rate
        :param arp_position: ARP position
        :param arp_velocity: ARP velocity
        :return: the intersection between the R/Rdot Contour and the DEM, if multiple intersections occur they will
                 be returned in order of increasing height above the WGS-84 ellipsoid.
        """

        # (1) Compute the center point (ctr) and the radius of the R/Rdot projection contour (rrrc).
        v_mag = np.linalg.norm(arp_velocity)
        u_vel = arp_velocity / v_mag
        cos_dca = -1.0 * r_dot_tgt_coa / v_mag
        sin_dca = np.sqrt(1 - cos_dca * cos_dca)
        ctr = arp_position + r_tgt_coa * cos_dca * u_vel
        rrrc = r_tgt_coa * sin_dca

        # (2) Compute the unit vectors u_rrx and u_rry to be used to compute points located on the R/Rdot contour.
        dec_arp = np.linalg.norm(arp_position)
        u_up = arp_position / dec_arp
        rry = np.cross(u_up, u_vel)
        u_rry = rry / np.linalg.norm(rry)
        u_rrx = np.cross(u_rry, u_vel)

        # (3) Compute the projection along the R/Rdot contour to the surface of constant HAE at height hae_max.
        # The projection point at height hae_max is point_a. Also compute the cosine and sine of the contour angle
        # to point_a, cos_caa and sin_caa.
        point_a = self.hae_max_surface_projection.rrdot_to_ground(r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity)[0]
        cos_caa = np.dot(point_a - ctr, u_rrx) / rrrc
        # This variable is defined in the specification but it does not appear to be used anywhere
        # sin_caa = self.look * np.sqrt(1 - cos_caa * cos_caa)

        # (4) Compute the projection along the R/Rdot contour to the surface of constant HAE at height hae_min.
        # The projection point at height hae_min is point_b. Also compute the cosine and sine of the contour angle
        # to point_b, cos_cab and sin_cab.
        point_b = self.hae_min_surface_projection.rrdot_to_ground(r_tgt_coa, r_dot_tgt_coa, arp_position, arp_velocity)[0]
        cos_cab = np.dot(point_b - ctr, u_rrx) / rrrc
        sin_cab = self.look * np.sqrt(1 - cos_cab * cos_cab)

        # (5) A set of points along the R/Rdot contour are to be computed. The points will be spaced in equal
        # increments of the cosine of the contour angle. Compute the step size, delta_cos_ca.

        # (5.1) Step size delta_cos_rrc is computed such that the distance between adjacent points on the R/Rdot
        # contour is approximately equal to delta_dist_rrc.
        delta_cos_rrc = self.delta_dist_rrc * np.abs(sin_cab) / rrrc

        # (5.2) Step size delta_cos_dem is computed such that the horizontal distance between adjacent points on
        # the R/Rdot contour is approximately equal to delta_dist_dem.
        delta_cos_dem = self.delta_dist_dem * (np.abs(sin_cab) / cos_cab) / rrrc

        # (5.3) Set delta_cos_ca (Note the value of delta_cos_ca is < 0)
        delta_cos_ca = -1.0 * min(delta_cos_rrc, delta_cos_dem)

        # (6) Determine the number of points along the R/Rdot contour to be computed, npts.
        npts = int(np.floor((cos_caa - cos_cab) / delta_cos_ca)) + 2

        # (7) Compute the set of points along the R/Rdot contour, {Pn} for n =0, 2, ..., npts-1. Initial point P1 is
        # located on the hae_min surface. The final point is located above the hae_max surface. Point Pn is computed
        # in ECF coordinates. Note that here n ranges from [0, npts-1] while in the specification n is [1, npts].
        # Equations have been modified accordingly.
        points_ecf = []
        for n in range(0, npts):
            cos_can = cos_cab + n * delta_cos_ca  # n-1 is unnecessary since n is zero based here
            sin_can = self.look * np.sqrt(1 - cos_can * cos_can)
            pn = ctr + rrrc * (cos_can * u_rrx + sin_can * u_rry)
            points_ecf.append(pn)

        # (8 - 10) For each of the NPTS points, convert from ECF coordinates to DEM coordinates (lon, lat, ele). Also
        # compute the DEM surface height for the point with DEM horizontal coordinates (lon, lat). Compute the
        # difference in height (delta_height) between the point on the contour and DEM surface point. Also,
        # set an indicator to track if that point is ABOVE, ON, or BELOW the DEM surface.
        #
        # Contour points that are within delta_hd_lim of the surface point are considered to be “on” the surface and
        # will be added to the result set. Also compute a result point when points n and n+1 when both are “off”
        # the surface and the R/Rdot contour intersects the surface between them (i.e. indicator n-1 x indicator n = -1)
        #
        # All height coordinates are in meters.
        intersection_points = []
        prev_indicator = None
        prev_delta_height = None
        for n in range(0, npts):
            point_lle = geocentric_to_geodetic(WorldCoordinate(points_ecf[n]))
            point_dem = GeodeticWorldCoordinate(point_lle.coordinate.copy())
            self.elevation_model.set_elevation(point_dem)
            delta_height = point_lle.elevation - point_dem.elevation

            # Determine if the contour point is ABOVE (indicator = 1), ON (indicator = 0), or BELOW (indicator = -1)
            if np.abs(delta_height) < self.delta_hd_lim:
                # Contour point is on the DEM surface, add it to the result set
                indicator = 0
                intersection_points.append(points_ecf[n])
            elif delta_height > self.delta_hd_lim:
                # Contour point is above the DEM surface
                indicator = 1
            else:
                # Contour point is below the DEM surface
                indicator = -1

            # If in two adjacent points one is ABOVE and the other BELOW then the contour intersected the DEM
            # surface between the two points. Interpolate an intersection point and add it to the result.
            if prev_indicator and prev_indicator * indicator == -1:
                # contour crossed between two points; compute the surface point between
                frac = prev_delta_height / (prev_delta_height - delta_height)
                cos_cas = cos_cab + (n - 1 + frac) * delta_cos_ca  # here the -1 is necessary for previous point
                sin_cas = self.look * np.sqrt(1 - cos_cas * cos_cas)
                sm = ctr + rrrc * (cos_cas * u_rrx + sin_cas * u_rry)
                intersection_points.append(sm)

            # Keep track of the current indicator and delta height for comparison to the next point
            prev_indicator = indicator
            prev_delta_height = delta_height

        # Return the set of surface points found. Points will be ordered of increasing height above the WGS-84
        # ellipsoid.
        return np.array(intersection_points)


class SICDSensorModel(SensorModel):
    """
    This is an implementation of the SICD sensor model as described by SICD Volume 3 Image Projections Description
    NGA.STND.0024-3_1.3.0 (2021-11-30)
    """

    def __init__(
        self,
        coord_converter: SARImageCoordConverter,
        coa_projection_set: COAProjectionSet,
        u_spn: np.ndarray,
        side_of_track: str = "L",
        u_gpn: Optional[np.ndarray] = None,
    ):
        """
        Constructs a SICD sensor model from the information derived from the XML metadata.

        :param coord_converter: converts coordinates between image grid and image plane
        :param coa_projection_set: projects image locations to the r/rdot contour
        :param u_spn: slant plane normal
        :param side_of_track: side of track imaged, "L" or "R", default "L"
        :param u_gpn: optional unit normal for ground plane
        """
        super().__init__()
        self.coa_projection_set = coa_projection_set
        self.coord_converter = coord_converter
        self.uvect_gpn = u_gpn
        self.uvect_spn = u_spn
        self.side_of_track = side_of_track

        self.default_surface_projection = GroundPlaneRRDotSurfaceProjection(self.coord_converter.scp_ecf, self.uvect_gpn)

    @staticmethod
    def compute_u_spn(scp_ecf: WorldCoordinate, scp_arp: np.ndarray, scp_varp: np.ndarray, side_of_track: str) -> np.ndarray:
        """
        This helper function computes the slant plane normal.

        :param scp_ecf: Scene Center Point position in ECF coordinates
        :param scp_arp: aperture reference point position
        :param scp_varp: aperture reference point velocity
        :param side_of_track: side of track imaged
        :return: unit vector for the slant plane normal
        """
        u_spn = np.cross(scp_varp, scp_ecf.coordinate - scp_arp)
        if side_of_track == "R":
            u_spn *= -1.0
        u_spn /= np.linalg.norm(u_spn)
        return u_spn

    @staticmethod
    def compute_u_gpn(scp_ecf: WorldCoordinate, u_row: np.ndarray, u_col: np.ndarray) -> np.ndarray:
        u_gpn = np.cross(u_row, u_col)
        u_gpn /= np.linalg.norm(u_gpn)
        if np.dot(u_gpn, scp_ecf.coordinate) < 0:
            u_gpn *= -1
        return u_gpn

    def image_to_world(
        self,
        image_coordinate: ImageCoordinate,
        elevation_model: Optional[ElevationModel] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> GeodeticWorldCoordinate:
        """
        This is an implementation of an Image Grid to Scene point projection that first projects the image
        location to the R/RDot contour and then intersects the R/RDot contour with the elevation model. If
        an elevation model is provided then this routine intersects the R/Rdot contour with the DEM surface which
        may result in multiple solutions. In that case the solution with the lowest HAE is returned.

        :param image_coordinate: the x,y image coordinate
        :param elevation_model: the optional elevation model, if none supplied a plane tangent to SCP is assumed
        :param options: no additional options are supported at this time
        :return: the lon, lat, elev geodetic coordinate of the surface matching the image coordinate
        """
        row_col = np.array(
            [
                image_coordinate.r + self.coord_converter.first_pixel.r,
                image_coordinate.c + self.coord_converter.first_pixel.c,
            ]
        )
        xrow_ycol = self.coord_converter.rowcol_to_xrowycol(row_col=row_col)
        r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa = self.coa_projection_set.precise_rrdot_computation(xrow_ycol)

        if elevation_model is not None:
            surface_projection = DEMRRDotSurfaceProjection(
                self.coord_converter.scp_ecf, self.side_of_track, elevation_model=elevation_model
            )
        else:
            surface_projection = self.default_surface_projection

        # Note that for a DEM the r/rdot contour may intersect the surface at multiple locations
        # resulting in an ambiguous location. Here we are arbitrarily selecting the first result.
        # TODO: Is there a better way to handle multiple DEM intersections?
        coords_ecf = surface_projection.rrdot_to_ground(r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa)

        return geocentric_to_geodetic(WorldCoordinate(coords_ecf[0]))

    def world_to_image(self, world_coordinate: GeodeticWorldCoordinate) -> ImageCoordinate:
        """
        This is an implementation of Section 6.1 Scene To Image Grid Projection for a single point.

        :param world_coordinate: lon, lat, elevation coordinate of the scene point
        :return: the x,y pixel location in this image
        """
        ecf_world_coordinate = geodetic_to_geocentric(world_coordinate)

        # TODO: Consider making these options like we have for image_to_world
        tolerance = 1e-2
        max_iterations = 10

        # (2) Ground plane points are projected along straight lines to the image plane to establish points.
        # The GP to IP projection direction is along the SCP COA slant plane normal. Also, compute the image
        # plane unit normal, uIPN. Compute projection scale factor SF as shown.
        uvect_proj = self.uvect_spn
        scale_factor = float(np.dot(uvect_proj, self.coord_converter.uvect_ipn))

        # (3) Set initial ground plane position G1 to the scene point position S.
        scene_point = np.array([ecf_world_coordinate.x, ecf_world_coordinate.y, ecf_world_coordinate.z])
        g_n = scene_point.copy()

        xrow_ycol_n = None
        cont = True
        iteration = 0
        while cont:
            iteration += 1

            # (4) Project ground plane point g_n to image plane point i_n. The projection distance is dist_n. Compute
            # image coordinates xrown and ycoln.
            dist_n = np.dot(self.coord_converter.scp_ecf.coordinate - g_n, self.coord_converter.uvect_ipn) / scale_factor
            i_n = g_n + dist_n * uvect_proj
            xrow_ycol_n = self.coord_converter.ipp_to_xrowycol(i_n)

            # (5) Compute the precise projection for image grid location (xrown, ycoln) to the ground plane containing
            # the scene point S. The result is point p_n. For image grid location (xrown, ycoln), compute COA
            # parameters per Section 2. Compute the precise R/Rdot projection contour per Section 4. Compute the
            # R/Rdot intersection with the ground plane per Section 5.
            r_tgt_coa, r_dot_tgt_coa, time_coa, arp_coa, varp_coa = self.coa_projection_set.precise_rrdot_computation(
                xrow_ycol_n
            )
            p_n = self.default_surface_projection.rrdot_to_ground(r_tgt_coa, r_dot_tgt_coa, arp_coa, varp_coa)

            # (6) Compute the displacement between ground plane point Pn and the scene point S.
            diff_n = scene_point - p_n[0]
            delta_gpn = np.linalg.norm(diff_n)
            g_n += diff_n

            # If the displacement is greater than the threshold (GP_MAX), compute point Gn+1 and repeat the
            # projections in steps (4) and (5) above. If the displacement is less than the threshold, accept image
            # grid location (xrown, ycoln) as the precise image grid location for scene point S.
            cont = delta_gpn > tolerance and (iteration < max_iterations)

        row_col = self.coord_converter.xrowycol_to_rowcol(xrow_ycol_n)

        # Convert the row_col image grid location to an x,y image coordinate. Note that row_col is in reference
        # to the full image, so we subtract off the first_pixel offset to make the image coordinate correct if this
        # is a chip.
        return ImageCoordinate(
            [row_col[1] - self.coord_converter.first_pixel.x, row_col[0] - self.coord_converter.first_pixel.y]
        )



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import logging
import math
from typing import List, Optional, Tuple, Union

import geojson
import numpy as np
import shapely
from scipy.interpolate import RectBivariateSpline

from aws.osml.photogrammetry import ElevationModel, GeodeticWorldCoordinate, ImageCoordinate, SensorModel

from .imaged_feature_property_accessor import ImagedFeaturePropertyAccessor


class LocationGridInterpolator:
    """
    This class can be used to approximate geodetic world coordinates from a grid of correspondences that is
    computed over a given area.
    """

    def __init__(
        self,
        sensor_model: SensorModel,
        elevation_model: Optional[ElevationModel],
        grid_area_ulx: float,
        grid_area_uly: float,
        grid_area_width: float,
        grid_area_height: float,
        grid_resolution: int,
    ) -> None:
        """
        Construct the grid of correspondences of the requested size/resolution from using the sensor model provided.

        :param sensor_model: SensorModel = the sensor model for the image
        :param elevation_model: Optional[Elevationmodel] = an optional external elevation model
        :param grid_area_ulx: float = the x component of the upper left corner of the grid in pixel space
        :param grid_area_uly: float = the y component of the upper left corner of the grid in pixel space
        :param grid_area_width: float = the width of the grid in pixels
        :param grid_area_height: float = the height of the grid in pixels
        :param grid_resolution: int = the number of points to calculate across the grid. Total points will be resolution^^2

        :return: None
        """
        xs = np.linspace(grid_area_ulx, grid_area_ulx + grid_area_width, grid_resolution)
        ys = np.linspace(grid_area_uly, grid_area_uly + grid_area_height, grid_resolution)
        longitude_values = np.empty(len(xs) * len(ys))
        latitude_values = np.empty(len(longitude_values))
        elevation_values = np.empty(len(longitude_values))
        i = 0
        for x in xs:
            for y in ys:
                world_coordinate = sensor_model.image_to_world(ImageCoordinate([x, y]), elevation_model=elevation_model)
                longitude_values[i] = world_coordinate.longitude
                latitude_values[i] = world_coordinate.latitude
                elevation_values[i] = world_coordinate.elevation
                i += 1
        longitude_values.shape = len(xs), len(ys)
        latitude_values.shape = longitude_values.shape
        elevation_values.shape = longitude_values.shape

        self.longitude_interpolator = RectBivariateSpline(xs, ys, longitude_values, kx=1, ky=1)
        self.latitude_interpolator = RectBivariateSpline(xs, ys, latitude_values, kx=1, ky=1)
        self.elevation_interpolator = RectBivariateSpline(xs, ys, elevation_values, kx=1, ky=1)
        self.elevation_model = elevation_model

    def __call__(self, *args, **kwargs):
        """
        Call this interpolation function given an image coordinate array.

        :param args: a single argument for the coordinate array
        :param kwargs: not used
        :return: a GeodeticWorldCoordinate for that image location
        """
        image_coord = args[0]
        world_coord = [
            self.longitude_interpolator(image_coord[0], image_coord[1])[0][0],
            self.latitude_interpolator(image_coord[0], image_coord[1])[0][0],
            self.elevation_interpolator(image_coord[0], image_coord[1])[0][0],
        ]
        world_coordinate = GeodeticWorldCoordinate(world_coord)
        if self.elevation_model is not None:
            self.elevation_model.set_elevation(world_coordinate)
        return world_coordinate


class Geolocator:
    """
    A Geolocator is a class that assign geographic coordinates for the features that are currently defined in image
    coordinates.
    """

    def __init__(
        self,
        property_accessor: ImagedFeaturePropertyAccessor,
        sensor_model: SensorModel,
        elevation_model: Optional[ElevationModel] = None,
        approximation_grid_size: int = 11,
    ) -> None:
        """
        Construct a geolocator given the context objects necessary for performing the calculations.

        :param property_accessor: facade used to access standard properties of an imaged feature
        :param sensor_model: sensor model for the image
        :param elevation_model: external elevation model
        :param approximation_grid_size: resolution of the approximation grid to use

        :return: None
        """
        self.sensor_model = sensor_model
        self.property_accessor = property_accessor
        self.elevation_model = elevation_model
        self.approximation_grid_size = approximation_grid_size

    def geolocate_features(self, features: List[geojson.Feature]) -> None:
        """
        Update the features to contain additional information from the context provided.

        :param features: List[geojson.Feature] = the input features to refine
        :return: None, the features are updated in place
        """

        if not features:
            return

        self._geolocate_features_using_approximation_grid(features)

    def _geolocate_features_using_approximation_grid(self, features: List[geojson.Feature]) -> None:
        """
        This method computes geolocations for features using an approximation grid. It is useful for dense feature
        sets where the cost of computing precise locations for a set of close features is expensive and unnecessary.
        Here we first compute a grid of locations covering the features in question that is then used to efficiently
        assign geolocations for all the features.

        :param features: List[geojson.Feature] = the input features
        :return: None, but the individual features have their geometry property updated
        """

        # Compute the boundary of these features. Normally this will be similar to the tile boundary but the
        # interpolation needs to be setup to cover the entire extent, so we calculate it explicitly here. If the
        # features happen to be very tightly packed and only occupy a small portion of the tile we will gain some
        # benefit by creating the same resolution of approximation grid over the smaller area.
        feature_bounds = [math.inf, math.inf, -math.inf, -math.inf]
        for feature in features:
            image_geometry = self.property_accessor.find_image_geometry(feature)
            if image_geometry is None:
                continue
            image_geometry_bbox = image_geometry.bounds
            feature_bounds[0] = min(feature_bounds[0], image_geometry_bbox[0])
            feature_bounds[1] = min(feature_bounds[1], image_geometry_bbox[1])
            feature_bounds[2] = max(feature_bounds[2], image_geometry_bbox[2])
            feature_bounds[3] = max(feature_bounds[3], image_geometry_bbox[3])

        # Expand the bounds to handle edge case where the features are all located at a single point or line.
        feature_bounds[0] -= 10
        feature_bounds[1] -= 10
        feature_bounds[2] += 10
        feature_bounds[3] += 10

        # Use the feature boundary to set up an approximation grid for the region
        grid_area_ulx = feature_bounds[0]
        grid_area_uly = feature_bounds[1]
        grid_area_width = feature_bounds[2] - feature_bounds[0]
        grid_area_height = feature_bounds[3] - feature_bounds[1]
        tile_interpolation_grid = LocationGridInterpolator(
            self.sensor_model,
            self.elevation_model,
            grid_area_ulx,
            grid_area_uly,
            grid_area_width,
            grid_area_height,
            self.approximation_grid_size,
        )

        for feature in features:
            # If the feature has the "imageBBox" property set then we will convert it to the "bbox" property defined
            # in the GeoJSON spec. This is a [minx, miny, maxx, maxy] bounds for this object where x is degrees
            # longitude and y is degrees latitude.
            image_bbox = self.property_accessor.get_image_bbox(feature)
            if image_bbox is not None:
                bbox = image_bbox.bounds
                center_xy = [
                    (bbox[0] + bbox[2]) / 2.0,
                    (bbox[1] + bbox[3]) / 2.0,
                ]
                bbox_corners_image_coords = [
                    [bbox[0], bbox[1]],
                    [bbox[0], bbox[3]],
                    [bbox[2], bbox[3]],
                    [bbox[2], bbox[1]],
                ]
                bbox_corners_world_coords = [
                    Geolocator.radians_coordinate_to_degrees(tile_interpolation_grid(corner))
                    for corner in bbox_corners_image_coords
                ]
                bbox_corners_world_geometry = shapely.MultiPoint(bbox_corners_world_coords)
                feature["bbox"] = bbox_corners_world_geometry.bounds

            # If the feature has the "imageGeometry" property set then we will convert it to the "geometry" property
            # defined in the GeoJSON spec. The "geometry" property will have the same type (e.g. Point, LineString,
            # Polygon) as the "imageGeometry" property. If the feature does not have the "imageGeometry" property
            # defined this
            image_geometry = self.property_accessor.get_image_geometry(feature)
            if image_bbox is None and image_geometry is None:
                logging.info(f"Feature may be using deprecated attributes: {feature}")
                image_geometry = self.property_accessor.find_image_geometry(feature)
                if image_geometry is None:
                    logging.warning(f"There isn't a valid detection shape for feature: {feature}")
                    continue

            if image_geometry is not None:
                center_xy = (image_geometry.centroid.x, image_geometry.centroid.y)
                feature["geometry"] = self._geolocate_image_geometry(image_geometry, tile_interpolation_grid)

            # Adding these because some visualization tools (e.g. kepler.gl) can perform more
            # advanced rendering (e.g. cluster layers) if the data points have single coordinates.
            center_location = tile_interpolation_grid(center_xy)
            feature["properties"]["center_longitude"] = math.degrees(center_location.longitude)
            feature["properties"]["center_latitude"] = math.degrees(center_location.latitude)

    @staticmethod
    def _geolocate_image_geometry(
        image_geometry: shapely.Geometry, interpolation_grid: LocationGridInterpolator
    ) -> Union[geojson.geometry.Geometry, geojson.GeometryCollection]:
        """
        This function converts a shape in image coordinates into a GeoJSON geometry using an interpolation grid.
        The image geometry, represented as a shapely Geometry object, is assumed to have coordinates in [x, y]
        pixel values. The resulting feature geometry, represented as a geojson Geometry object, will have coordinates
        in [longitude, latitude, elevation] where longitude and latitude are in degrees and elevation is in meters.

        :param image_geometry: the image shape
        :param interpolation_grid: the interpolation grid derived from the sensor model
        :return: a GeoJSON geometry object
        """
        if image_geometry is None:
            raise ValueError("Unable to geolocate features without a geometry property")

        if isinstance(image_geometry, shapely.Point):
            print(f"Geolocating point: {image_geometry.coords[0]}")
            return geojson.Point(Geolocator.radians_coordinate_to_degrees(interpolation_grid(image_geometry.coords[0])))
        elif isinstance(image_geometry, (shapely.LineString, shapely.LinearRing)):
            return geojson.LineString(
                [Geolocator.radians_coordinate_to_degrees(interpolation_grid(coord)) for coord in image_geometry.coords]
            )
        elif isinstance(image_geometry, shapely.Polygon):
            boundary_coords = [
                [
                    Geolocator.radians_coordinate_to_degrees(interpolation_grid(coord))
                    for coord in image_geometry.exterior.coords
                ]
            ]
            for i in image_geometry.interiors:
                boundary_coords.append(
                    [Geolocator.radians_coordinate_to_degrees(interpolation_grid(coord)) for coord in i.coords]
                )
            return geojson.Polygon(boundary_coords)
        elif (
            image_geometry.__class__.__name__.startswith("Multi")
            or image_geometry.__class__.__name__ == "GeometryCollection"
        ):
            geometry_list = [Geolocator._geolocate_image_geometry(part, interpolation_grid) for part in image_geometry.geoms]
            print(f"geometry_list: {geometry_list}")
            return getattr(geojson, image_geometry.__class__.__name__)(geometry_list)
        else:
            raise ValueError(f"Unhandled geometry type: {image_geometry.__class__}")

    @staticmethod
    def radians_coordinate_to_degrees(
        coordinate: GeodeticWorldCoordinate,
    ) -> Tuple[float, float, float]:
        """
        GeoJSON coordinate order is a decimal longitude, latitude with an optional height as a 3rd value
        (i.e. [lon, lat, ht]). The WorldCoordinate uses the same ordering but the longitude and latitude are expressed
        in radians rather than degrees.

        :param coordinate: GeodeticWorldCoordinate = the geodetic world coordinate (longitude, latitude, elevation)

        :return: Tuple[float, float, float] = degrees(longitude), degrees(latitude), elevation
        """
        return (
            math.degrees(coordinate.longitude),
            math.degrees(coordinate.latitude),
            coordinate.elevation,
        )
